import math

import numpy as np
import pandas as pd

from fedex_generator.commons import utils
from fedex_generator.Measures.BaseMeasure import BaseMeasure, START_BOLD, END_BOLD
from fedex_generator.Measures.Bins import Bin


class ExceptionalityMeasure(BaseMeasure):
    """
    A class implementing the Exceptionality measure, a measure for interestingness of filter / join / union operations.\n
    This measure is described in the article "FEDEX: An Explainability Framework for Data Exploration Steps" by
    Daniel Deutch, Amir Gilad, Tova Milo, Amit Mualem, Amit Somech.\n
    From the article: "this measure intuitively deems the filter result interesting if it produces an output dataframe
    which significantly deviates from the input dataframe."\n
    This measure is based on the Kolmogorov-Smirnov test, a well known statistical test used to assess whether two
    samples originate from the same probability distribution.\n
    The measure is defined as: \n
    .. math:: I_A (d_{in}, q_f, d_{out}) = KS(\\Pr(d_{in}), \\Pr(d_{out}))\n
    Where:\n
    - :math:`I_A` is the interestingness of column A.\n
    - :math:`d_{in}` is the input dataframe.\n
    - :math:`q_f` is the query operation.\n
    - :math:`d_{out}` is the output dataframe.\n
    - :math:`KS` is the Kolmogorov-Smirnov test.\n
    - :math:`\\Pr(d_{in})` is the probability distribution of the input dataframe.\n
    - :math:`\\Pr(d_{out})` is the probability distribution of the output dataframe.\n
    """

    def __init__(self):
        super().__init__()

    def draw_bar(self, bin_item: Bin, influence_vals: dict = None, title=None, ax=None, score=None,
                 show_scores: bool = False, explanation_num: int | None = None) -> str:
        """
        Draw a bar chart comparing the distribution of values before and after an operation.

        This method generates a bar chart that visualizes the distribution of values in a bin before and after
        an operation. It highlights the most influential value in green if influence values are provided.

        :param bin_item: The bin object containing the data to be visualized.
        :param influence_vals: A dictionary of influence values for each bin value (default is None).
        :param title: The title of the bar chart (default is None).
        :param ax: The matplotlib axes object to draw the bar chart on (default is None).
        :param score: The score to be displayed in the title if show_scores is True (default is None).
        :param show_scores: A boolean indicating whether to display the score in the title (default is False).

        :return: The name of the bin.
        """
        # Get the source and result columns
        res_col = bin_item.get_binned_result_column()
        src_col = bin_item.get_binned_source_column()

        # Get the probabilities of the source and result columns
        res_probs = res_col.value_counts(normalize=True)
        src_probs = None if src_col is None else src_col.value_counts(normalize=True)
        labels = set(src_probs.keys()).union(res_probs.keys())

        # Get the 25 most influential values if there are more than 25 values
        MAX_BARS = 25
        if len(labels) > MAX_BARS:
            labels, _ = self.get_max_k(influence_vals, MAX_BARS)

        labels = sorted(labels)
        probabilities = [100. * src_probs.get(item, 0) for item in labels]
        probabilities2 = [100 * res_probs.get(item, 0) for item in labels]

        width = 0.35
        ind = np.arange(len(labels))

        # Draw the bar chart
        result_bar = ax.bar(ind + width, probabilities2, width, label="After Filter")

        ax.bar(ind, probabilities, width, label="Before Filter")
        ax.legend(loc='best')

        # Highlight the most influential value in green, if influence values are provided
        if influence_vals:
            max_label, _ = self.get_max_k(influence_vals, 1)
            max_label = max_label[0]
            result_bar[labels.index(max_label)].set_color('green')

        # Set the labels and title of the bar chart
        ax.set_xticks(ind + width / 2)
        label_tags = tuple([utils.to_valid_latex(bin_item.get_bin_representation(i)) for i in labels])
        tags_max_length = max([len(tag) for tag in label_tags])
        ax.set_xticklabels(label_tags, rotation='vertical' if tags_max_length >= 4 else 'horizontal')

        ax.set_xlabel(utils.to_valid_latex(bin_item.get_bin_name() + " values"), fontsize=20)
        ax.set_ylabel("frequency(\\%)", fontsize=16)

        # Set the title of the bar chart. If show_scores is True, display the score in the title
        if explanation_num is not None:
            title = f"{START_BOLD}[{explanation_num}]{END_BOLD} {title}" if title is not None else f"{START_BOLD}[{explanation_num}]{END_BOLD}"
        if title is not None:
            if show_scores:
                ax.set_title(f'score: {score}\n {utils.to_valid_latex(title)}', fontdict={'fontsize': 20})
            else:
                ax.set_title(utils.to_valid_latex(title), fontdict={'fontsize': 20})

        # Rotate the x-axis labels by 45 degrees if there are more than 8 labels
        if len(labels) > 8:
            ax.tick_params(axis='x', rotation=45, tickdir='in')

        ax.set_axis_on()
        return bin_item.get_bin_name()  ####

    def interestingness_only_explanation(self, source_col, result_col, col_name):
        # If the column is categorical, compare the most common value before and after the operation
        if utils.is_categorical(source_col):
            vc = source_col.value_counts()
            source_max = utils.max_key(vc)
            vc = result_col.value_counts()
            result_max = utils.max_key(vc)
            return f"The distribution of column '{col_name}' changed significantly.\n" \
                   f"The most common value was {source_max} and now it is {result_max}."

        # Otherwise, compare the mean and standard deviation of the column before and after the operation
        std_source = np.sqrt(np.var(source_col))
        mean_source = np.mean(source_col)
        std = np.sqrt(np.var(result_col))
        mean = np.mean(result_col)

        return f"The distribution of column '{col_name}' changed significantly.\n" \
               f" The mean was {mean_source:.2f} and the standard " \
               f"deviation was {std_source:.2f}, and now the mean is {mean:.2f} and the standard deviation is {std:.2f}."

    def calc_measure_internal(self, bin: Bin):
        return ExceptionalityMeasure.kstest(bin.source_column.dropna(),
                                            bin.result_column.dropna())  # / len(source_col.dropna().value_counts())

    @staticmethod
    def kstest(s, r) -> float:
        """
        Calculate the Kolmogorov-Smirnov statistic between two samples.
        Uses the :func:`commons.kstest.ks_2samp` function to calculate the statistic.
        :param s: The first sample.
        :param r: The second sample.
        :return: The Kolmogorov-Smirnov statistic between the two samples.
        """
        s = np.array(s)
        s = s[s == s]
        r = np.array(r)
        r = r[r == r]
        return 0 if (len(r) == 0 or len(s) == 0) else utils.ks_2samp(s, r).statistic

    def calc_influence_col(self, current_bin: Bin):
        # Get the source and result columns
        bin_values = current_bin.get_bin_values()
        source_col = current_bin.get_source_by_values(bin_values)
        res_col = current_bin.get_result_by_values(bin_values)

        # Truncate the bin values to the 15 most common values
        if len(bin_values) > 15:
            bin_values = list(source_col.value_counts().nlargest(15).keys())

        # Get the measure score for the entire bin
        score_all = ExceptionalityMeasure.kstest(source_col, res_col)

        # Compute the influence of each value in the bin, as the difference between the score of the entire bin and the
        # score of the bin without the value
        influence = []
        for value in bin_values:
            source_col_only_list = current_bin.get_source_by_values([b for b in bin_values if b != value])
            res_col_only_list = current_bin.get_result_by_values([b for b in bin_values if b != value])

            score_without_bin = ExceptionalityMeasure.kstest(source_col_only_list, res_col_only_list)
            influence.append(score_all - score_without_bin)

        return influence

    def build_operation_expression(self, source_name):
        from fedex_generator.Operations.Filter import Filter
        from fedex_generator.Operations.Join import Join

        if isinstance(self.operation_object, Filter):
            if self.operation_object.value is not None and self.operation_object.operation_str is not None:
                return f'Dataframe {self.operation_object.source_name}, ' \
                       f'filtered on attribute {self.operation_object.attribute} ' \
                       f'{self.operation_object.operation_str} {self.operation_object.value}'
            else:
                return f'Dataframe {self.operation_object.source_name}, ' \
                       f'filtered on attribute {self.operation_object.attribute}'
        elif isinstance(self.operation_object, Join):
            return f'{self.operation_object.right_name} joined with {self.operation_object.left_name} by {self.operation_object.attribute}'

    def build_explanation(self, current_bin: Bin, col_name, max_value, source_name):
        # Get the source and result columns, then calculate the probabilities of each value in the columns
        source_col = current_bin.get_binned_source_column()
        res_col = current_bin.get_binned_result_column()

        res_probs = res_col.value_counts(normalize=True)
        source_probs = source_col.value_counts(normalize=True)

        # Add the missing values to the dictionaries with a probability of 0
        for bin_value in current_bin.get_bin_values():
            res_probs[bin_value] = res_probs.get(bin_value, 0)
            source_probs[bin_value] = source_probs.get(bin_value, 0)

        additional_explanation = []
        if current_bin.name == "NumericBin":
            # Get the index of the most influential value
            values = current_bin.get_bin_values()
            index = values.index(max_value)

            values_range_str = "below {}".format(utils.format_bin_item(values[1])) if max_value == 0 else \
                "above {}".format(utils.format_bin_item(max_value)) if index == len(values) - 1 else \
                    "between {} and {}".format(utils.format_bin_item(values[index]),
                                               utils.format_bin_item(values[index + 1]))

            # Compute the factor - the ratio between the probability of the value in the result column and the
            # probability of the value in the source column
            factor = res_probs.get(max_value, 0) / source_probs[max_value]
            proportion = "less" if factor < 1 else "more"

            # If the factor is between 0 and 1, then the value now appears less frequently, and we set the factor
            # to 1 / factor for a more intuitive explanation
            # (e.g. appears 3 times less frequently instead of 0.33 times)
            if (factor < 1 and factor > 0):
                factor = 1 / factor

            # If the factor is 0, the value no longer exists in the result column
            if factor == 0:
                additional_explanation.append(
                    f"{START_BOLD}{utils.to_valid_latex(col_name, True)}{END_BOLD} values "
                    f"{START_BOLD}{utils.to_valid_latex(values_range_str, True)}{END_BOLD}\n"
                    f"are {START_BOLD}no{END_BOLD} {START_BOLD}longer{END_BOLD} {START_BOLD}exist{END_BOLD} (was {round(source_probs[max_value], 3) * 100}%)")
            else:
                appear_test = f'{utils.smart_round(factor)} times {proportion}'
                additional_explanation.append(
                    f"{START_BOLD}{utils.to_valid_latex(col_name, True)}{END_BOLD} values "
                    f"{START_BOLD}{utils.to_valid_latex(values_range_str, True)}{END_BOLD} (in green)\n"
                    f"appear {START_BOLD}"
                    f"{utils.to_valid_latex(appear_test, True)}"
                    f"{END_BOLD} than before")
        else:
            factor = res_probs.get(max_value, 0) / source_probs[max_value]

            source_prob = 100 * source_probs[max_value]
            res_prob = 100 * res_probs.get(max_value, 0)
            max_value_rep = current_bin.get_bin_representation(max_value)

            if factor == 0:
                proportion_sentes = f' frequency was {source_prob:.1f}% now {res_prob:.1f}%'
            elif factor < 1:
                proportion = "less"
                factor = 1 / factor
                proportion_sentes = f"appear {START_BOLD} " \
                                    f"{utils.to_valid_latex(f'{utils.smart_round(factor)} times {proportion}', True)}" \
                                    f" {END_BOLD} than before"
            else:
                proportion = "more"
                proportion_sentes = f"appear {START_BOLD} " \
                                    f"{utils.to_valid_latex(f'{utils.smart_round(factor)} times {proportion}', True)}" \
                                    f" {END_BOLD} than before"

            additional_explanation.append(
                f"{START_BOLD}{utils.to_valid_latex(col_name, True)}{END_BOLD} value "
                f"{START_BOLD}{utils.to_valid_latex(max_value_rep, True)}"
                f"{END_BOLD} (in green)\n{proportion_sentes}")

        influence_top_example = ", ".join(additional_explanation)

        return influence_top_example



import numpy as np

from fedex_generator.Measures.DiversityMeasure import DiversityMeasure
from fedex_generator.commons import utils


class NormalizedDiversityMeasure(DiversityMeasure):
    """
    A normalized version of the diversity measure.\n
    See the DiversityMeasure class for more information.\n
    The only difference is in the calculation of the variance. The square root of the variance is divided by the mean
    of the appearances of the values in the array, where as in the DiversityMeasure class, this is not done.
    """

    def __init__(self):
        super().__init__()

    def calc_var(self, pd_array):
        if utils.is_numeric(pd_array):
            return np.abs(np.nanstd(pd_array) / np.nanmean(pd_array))

        appearances = (pd_array.value_counts()).to_numpy()
        mean = np.mean(appearances)
        variance = np.sum(np.power(appearances - mean, 2)) / len(appearances)
        print(variance, np.sqrt(variance) / mean)

        return np.sqrt(variance) / mean



from pandas import Series


class DatasetRelation(object):
    """
    A class that represents the relation between the source and the result of an operation.
    """

    def __init__(self, source_df, result_df, source_table_name):
        """
        :param source_df: The source DataFrame
        :param result_df: The result DataFrame
        :param source_table_name: The name of the source table
        """
        self.source_df = source_df
        self.result_df = result_df
        self.source_table_name = source_table_name

    def get_source(self, attr) -> Series | None:
        """
        Get an attribute from the source DataFrame
        """
        # GroupBy
        if self.source_df is None or attr not in self.source_df:
            return None
        return self.source_df[attr]

    def get_result(self, attr) -> Series | None:
        """
        Get an attribute from the result DataFrame
        """
        return self.result_df[attr]

    def get_source_name(self) -> str:
        """
        Get the name of the source table
        """
        return self.source_table_name



import json

import numpy.random
import pandas as pd
from extract_intermediate_results import run_on_all_queries, Query, set_working_directory
import sys
import os
from typing import List
import io
from concurrent.futures import ThreadPoolExecutor
import warnings
import random

random.seed(42)
numpy.random.seed(42)
import test_utils as tu
import test_consts as consts

warnings.filterwarnings("ignore")

default_results_path = "resources/results_files"


def load_all_results(paths: List[str] | None = None) -> List[dict]:
    """
    Load all result files from the given paths.
    If no paths are given, all files from the default results directory are loaded.
    If paths are provided, it must be a list of paths to the result files (and not to directories).
    :param paths: List of paths to result files. Leave None to load all files from the default results directory.
    :return: List of dictionaries, each containing the content of a result file.
    """
    if paths is None:
        paths = [f"{default_results_path}/{file}" for file in os.listdir(default_results_path)]
    results = []
    for path in paths:
        with open(path, "r") as file:
            results.append(json.load(file))
        file.close()

    # Convert the dataframes that were converted to strings back to dataframes.
    for result in results:
        for k, v in result.items():
            # Try converting the key to a query. If this works, the value is a dict.
            # Otherwise, it's one of the metadata fields, like "dataset_name" or "dataset_path".
            try:
                query = Query.from_string(k)
                v['query_object'] = query
            # Index error occurs when the key is not a query.
            except IndexError:
                continue
            if "saved_results" in v:
                # Wrapping the string in a StringIO object to be able to read it as a file.
                result[k]["saved_results"] = pd.read_json(io.StringIO(v["saved_results"]))
            if "final_df" in v and isinstance(v["final_df"], str):
                result[k]["final_df"] = pd.read_json(io.StringIO(v["final_df"]))

    return results


def compare_results(result_files: List[dict], re_produced_results: List[dict]):
    failed_on_datasets = []
    passed_on_datasets = []
    for result in result_files:
        print(f"\n \n -------------------------------------------------- \n"
              f"\033[1m Comparing results on dataset {result['dataset_name']} \033[0;0m \n --------------------------------------------------")
        matching_result = [r for r in re_produced_results if r['idx'] == result['idx']][0]
        test_outcomes = []
        for k, v in result.items():
            if not isinstance(v, dict):
                continue
            test_outcomes.append({
                'test_name': k,
                'passed': True,
                'failed_tests': []
            })
            print(
                f"\n  \n -------------------------------------------------- \n Comparing for query {k} \n --------------------------------------------------")
            matching_result_query = matching_result[k]

            for test_name, test_args in consts.test_funcs.items():
                # Get the attribute name to test, if it exists in the result.
                test_attribute = test_args['attribute_name']
                if test_attribute not in v:
                    continue
                # It may be required to rename duplicate columns or bins, if they exist (otherwise the test may
                # fail because of a comparison between two different columns with the same name).
                require_duplicate_fix = test_args['require_duplicate_fix']
                if require_duplicate_fix:
                    matching_result_query[test_attribute] = tu.fix_duplicate_col_names_and_bin_names(
                        matching_result_query[test_attribute])
                    v[test_attribute] = tu.fix_duplicate_col_names_and_bin_names(v[test_attribute])

                # Run the test and print the results.
                test_passed, errors = test_args['func'](v[test_attribute], matching_result_query[test_attribute])
                tu.print_test_messages(errors, test_name, test_passed)
                if not test_passed:
                    test_outcomes[-1]['passed'] = False
                    test_outcomes[-1]['failed_tests'].append(test_name)

        tu.print_result_summary(test_outcomes, result['dataset_name'])
        if not all([outcome['passed'] for outcome in test_outcomes]):
            failed_on_datasets.append(result['dataset_name'])
        else:
            passed_on_datasets.append(result['dataset_name'])

    tu.print_execution_summary(failed_on_datasets, passed_on_datasets)


def main(argv):
    set_working_directory()

    result_files = load_all_results(argv[1:] if len(argv) > 1 else None)

    for i in range(len(result_files)):
        result_files[i]['idx'] = i

    datasets = tu.load_datasets(result_files)

    # If a global column select is present in any result file, apply it to the relevant dataset, and save
    # a separate copy of the dataset with the global select applied and an identifier of the specific result.
    for result in result_files:
        if 'global_select' in result:
            dataset = datasets[result['dataset_name']]
            global_select = result['global_select']
            dataset_with_global_select = dataset[global_select]
            datasets[f"{result['dataset_name']}_{result['idx']}"] = dataset_with_global_select
            result['dataset_name'] = f"{result['dataset_name']}_{result['idx']}"
        if 'global_select_second' in result:
            dataset = datasets[result['second_dataset_name']]
            global_select = result['global_select_second']
            dataset_with_global_select = dataset[global_select]
            datasets[f"{result['second_dataset_name']}_{result['idx']}"] = dataset_with_global_select
            result['second_dataset_name'] = f"{result['second_dataset_name']}_{result['idx']}"

    result_queries = [
        [v['query_object'] for k, v in result.items() if isinstance(v, dict) and 'query_object' in v]
        for result in result_files
    ]

    print(
        "Reproducing results using the current implementation. Please be patient, this may take a while depending on the number of queries and datasets.\n")
    # Use a thread pool to run the queries in parallel.
    with ThreadPoolExecutor() as executor:
        re_produced_results = list(executor.map(
            run_on_all_queries,
            result_queries,
            [datasets[result['dataset_name']] for result in result_files],
            [datasets[result['second_dataset_name']] for result in result_files],
            range(len(result_files)),
            [False] * len(result_files)
        ))

    compare_results(result_files, re_produced_results)


if __name__ == '__main__':
    main(sys.argv)



from typing import List, Dict
from colorama import Fore
from colorama import init as colorama_init
import test_consts as consts
import pandas as pd
from pandas import DataFrame
import json
import os

colorama_init(autoreset=True)


def save_dataset_to_resources(dataset: DataFrame, dataset_name: str):
    """
    Save the dataset to resources/datasets.
    :param dataset: The dataset to save.
    :param dataset_name: The name of the dataset.
    """
    # Check if resources/datasets exists. If not, create it.
    if not os.path.exists("resources/datasets"):
        os.makedirs("resources/datasets")
    dataset.to_csv(f"resources/datasets/{dataset_name}.csv", index=False)


def load_dataset(path: str | None, dataset_name: str, default_datasets: List[Dict]) -> DataFrame | None:
    """
    Load a dataset from the given path.\n
    If the dataset can not be loaded from the given path, we try loading it by name from resources/datasets, the
    default library.
    If this too fails, we try loading it from the default datasets configuration by downloading it from the internet.
    :param path: Path to the dataset.
    :param dataset_name: Name of the dataset.
    :param default_datasets: List of dictionaries, each containing the name and path of a default dataset.
    :return: The loaded dataset.
    """
    if path is None:
        return None
    try:
        return pd.read_csv(path)
    except FileNotFoundError:
        try:
            return pd.read_csv(f"resources/datasets/{dataset_name}.csv")
        except FileNotFoundError:
            for default_dataset in default_datasets:
                if default_dataset['name'] == dataset_name:
                    dataset = pd.read_csv(default_dataset['link'])
                    # Save the dataset to resources/datasets for future use.
                    save_dataset_to_resources(dataset, dataset_name)
                    return dataset
        raise FileNotFoundError(f"Dataset not found at path {path} and {dataset_name} not found in default datasets.")


def load_datasets(results: List[dict], default_configuration_path: str = "resources/default_datasets.json") -> Dict[
    str, DataFrame]:
    """
    Load all datasets from the given results.
    If a dataset is present in multiple results, only the first occurrence is considered.
    If a dataset can not be loaded but its name is present in the default configuration, the dataset will be loaded
    using the default configuration.
    :param results: List of dictionaries, each containing the content of a result file, loaded by load_all_results.
    :return: A dictionary with dataset names as keys and DataFrames as values.
    """
    default_datasets = list(json.load(open(default_configuration_path, "r"))['all'])
    datasets = {}

    for result in results:
        # Get the dataset name and path from the result.
        first_dataset_path = result['first_dataset']
        first_dataset_name = result['dataset_name']
        second_dataset_path = result['second_dataset']
        second_dataset_name = result['second_dataset_name']

        # Load the first dataset.
        if first_dataset_name not in datasets:
            datasets[first_dataset_name] = load_dataset(first_dataset_path, first_dataset_name, default_datasets)
        # Load the second dataset.
        if second_dataset_name not in datasets:
            datasets[second_dataset_name] = load_dataset(second_dataset_path, second_dataset_name, default_datasets)

    return datasets


def fix_duplicate_col_names_and_bin_names(dict_list: List[Dict]):
    """
    This function takes a list of influence values dictionaries or significance values dictionaries,
    and fixes the duplicate column names and bin names by appending a number to the column name and bin name.
    :param dict_list: List of dictionaries, in the format of influence values or significance values.
    :return: List of dictionaries, with fixed duplicate column names and bin names.
    """
    keys = set()
    idx_dict = {}
    for d in dict_list:
        col_name = d['column']
        bin_name = d['bin']
        if col_name in keys:
            idx_dict[col_name] += 1
            d['column'] = f'{col_name}_{idx_dict[col_name]}'
            d['bin'] = f'{bin_name}_{idx_dict[col_name]}'
        else:
            keys.add(col_name)
            idx_dict[col_name] = 1
    return dict_list


def print_test_messages(messages: List[str], test_name: str, passed: bool) -> None:
    """
    This function prints the test messages, the test name and whether the test passed or not.
    :param messages: List of messages to print.
    :param test_name: Name of the test.
    :param passed: Boolean value, whether the test passed or not.
    :return: None
    """
    if passed:
        color = Fore.GREEN
    else:
        color = Fore.RED

    print(f"\n{color} --------------------------------------------------------------")
    print(f"\t{color} {test_name}:\n")

    for message in messages:
        print(f"\t\t{color} {message}")

    if passed:
        print(f"{'\n' if len(messages) > 0 else ''}\t{color} TEST PASSED")
    else:
        print(f"{'\n' if len(messages) > 0 else ''}\t{color} TEST FAILED")
    print(f"{color} --------------------------------------------------------------")


def print_result_summary(test_outcomes: List[Dict], dataset_name: str) -> None:
    """
    This function prints the summary of the test outcomes.
    If any test failed, the function will print the failed tests.
    :param test_outcomes: List of dictionaries, each containing the test name, whether the test passed or not, and the failed tests.
    :param dataset_name: Name of the dataset.
    :return: None
    """
    passed_tests = [outcome for outcome in test_outcomes if outcome['passed']]
    failed_tests = [outcome for outcome in test_outcomes if not outcome['passed']]

    print(f"\n\n{Fore.CYAN} --------------------------------------------------------------")
    print(f"\t{Fore.CYAN} SUMMARY OF TESTS ON DATASET {dataset_name}:\n")
    print(f"\t{Fore.CYAN} QUERIES THAT PASSED ALL TESTS: {len(passed_tests)}")
    print(f"\t{Fore.CYAN} QUERIES WITH FAILED TESTS: {len(failed_tests)}")

    if len(passed_tests) > 0:
        print(f"\n\t{Fore.GREEN} PASSED ALL TESTS ON QUERIES:")
        for test in passed_tests:
            print(f"\t\t{Fore.GREEN}- {test['test_name']}")

    if len(failed_tests) > 0:
        print(f"\n\t{Fore.RED} FAILED TESTS ON QUERIES:")
        for test in failed_tests:
            print(f"\t\t{Fore.RED}- {test['test_name']}")
            print(f"\t\t{Fore.RED} TESTS FAILED ON THIS QUERY:")
            for failed_test in test['failed_tests']:
                print(f"\t\t\t{Fore.RED}- {consts.test_fail_explanations[failed_test]}")
        print(f"\n\t{Fore.RED} See above test results for more details on the failed tests.")
    else:
        print(f"\n\t{Fore.GREEN} ALL TESTS PASSED ON DATASET!")

    print(f"{Fore.CYAN} --------------------------------------------------------------")


def print_execution_summary(failed_on_datasets: List[str], passed_on_datasets: List[str]) -> None:
    """
    This function prints the summary of the test execution.
    It prints which datasets had every test pass, and which datasets had failed tests.
    :param failed_on_datasets: List of datasets that had failed tests.
    :param passed_on_datasets: List of datasets that had every test pass.
    :return: None
    """
    print(f"\n\n{Fore.CYAN} --------------------------------------------------------------")
    print(f"\t{Fore.CYAN} SUMMARY OF TEST EXECUTION:")
    print(f"\t{Fore.CYAN} DATASETS THAT PASSED ALL TESTS: {len(passed_on_datasets)}")
    print(f"\t{Fore.CYAN} DATASETS WITH FAILED TESTS: {len(failed_on_datasets)}")

    if len(passed_on_datasets) > 0:
        print(f"\n\t{Fore.GREEN} PASSED ALL TESTS ON DATASETS:")
        for dataset in passed_on_datasets:
            print(f"\t\t{Fore.GREEN}- {dataset}")

    if len(failed_on_datasets) > 0:
        print(f"\n\t{Fore.RED} FAILED TESTS ON DATASETS:")
        for dataset in failed_on_datasets:
            print(f"\t\t{Fore.RED}- {dataset}")
        print(f"\n\t{Fore.RED} See above test results for more details on the failed tests.")
    else:
        print(f"\n\t{Fore.GREEN} ALL TESTS PASSED ON ALL DATASETS!")
    print(f"{Fore.CYAN} --------------------------------------------------------------")



import pandas as pd

from fedex_generator.commons.consts import TOP_K_DEFAULT, DEFAULT_FIGS_IN_ROW
from fedex_generator.commons import utils
from fedex_generator.commons.DatasetRelation import DatasetRelation
from fedex_generator.Operations import Operation
from fedex_generator.Measures.ExceptionalityMeasure import ExceptionalityMeasure
from fedex_generator.Measures.ShapleyMeasure import ShapleyMeasure

def get_df_name(df, scope):
    name =[x for x in scope if scope[x] is df][0]
    return name
class BJoin(Operation.Operation):
    def __init__(self, left_df, right_df, source_scheme, attribute, result_df=None, left_name=None, right_name=None):
        super().__init__(source_scheme)
        self.source_scheme = source_scheme
        self.attribute = attribute

        df_join = pd.merge(left_df, right_df, on=attribute, how='inner')
            # for a, r, v in cond:
            #     if r == '=':
            #         df_join = df_join[df_join[a] == v]
            #     elif r == '>':
            #         df_join = df_join[df_join[a] > v]
            #     elif r == '>':
            #         df_join = df_join[df_join[a] < v]
        if df_join.shape[0]:
            self.result = True
        else: self.result = False
        self.left_name = utils.get_calling_params_name(left_df)
        self.right_name = utils.get_calling_params_name(right_df)
        self.left_df = left_df
        self.right_df = right_df
        # self.result_df = result_df

    def iterate_attributes(self):
        for attr in self.left_df.columns:
            if attr.lower() == "index":
                continue
            yield attr, DatasetRelation(self.left_df, self.result_df, self.left_name)

        for attr in set(self.right_df.columns) - set(self.left_df.columns):
            if attr.lower() == "index":
                continue
            yield attr, DatasetRelation(self.right_df, self.result_df, self.right_name)

    def explain(self, schema=None, attributes=None, top_k=TOP_K_DEFAULT,
                figs_in_row: int = DEFAULT_FIGS_IN_ROW, show_scores: bool = False, title: str = None, corr_TH: float = 0.7, consider='left'):

        if attributes is None:
            attributes = []

        if schema is None:
            schema = {}
        measure = ShapleyMeasure()
        if self.result:
            score = 1
        else: score = 0
        top_fact, facts = measure.get_most_contributing_fact(self, self.left_df, self.right_df, self.attribute,None, consider=consider)
        exp = f'The result of joining dataframes \'{self.left_name}\' and \'{self.right_name}\' on attribute \'{self.attribute}\' is not empty.\nThe following fact from dataframe \'{self.left_name}\' has significantly contributed to this result:\n'
        facts = list({k: v for k, v in sorted(facts.items(), key=lambda item: -item[1])}.items())
        fact_idx = 0
        exp += str(facts[fact_idx][0])
        top_k -= 1
        while top_k > 0:
            fact_idx += 1
            exp += '\n and then\n'
            exp += str(facts[fact_idx][0])
            top_k -= 1
        return 



import math

from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from itertools import combinations
import math
import inspect

from fedex_generator.commons import utils
from fedex_generator.Measures.BaseMeasure import BaseMeasure, START_BOLD, END_BOLD
from fedex_generator.Measures.Bins import Bin


def is_name_informative(name):
    import string
    return any([char in string.ascii_letters for char in name])

def get_calling_params_name(item):
    frames = inspect.stack()
    highest_var = None
    for frame in frames[2:]:
        prev_locals = frame[0].f_locals
        for var_name in prev_locals:
            if id(item) == id(prev_locals[var_name]) and is_name_informative(var_name):
                highest_var = var_name
    return highest_var


class ShapleyMeasure(BaseMeasure):
    def __init__(self):
        super().__init__()




    def draw_bar(self, bin_item: Bin, influence_vals: dict = None, title=None, ax=None, score=None,
                 show_scores: bool = False):
        res_col = bin_item.get_binned_result_column()
        src_col = bin_item.get_binned_source_column()

        res_probs = res_col.value_counts(normalize=True)
        src_probs = None if src_col is None else src_col.value_counts(normalize=True)
        labels = set(src_probs.keys()).union(res_probs.keys())

        MAX_BARS = 25
        if len(labels) > MAX_BARS:
            labels, _ = self.get_max_k(influence_vals, MAX_BARS)

        labels = sorted(labels)
        probabilities = [100. * src_probs.get(item, 0) for item in labels]
        probabilities2 = [100 * res_probs.get(item, 0) for item in labels]

        width = 0.35
        ind = np.arange(len(labels))

        result_bar = ax.bar(ind + width, probabilities2, width, label="After")

        ax.bar(ind, probabilities, width, label="Before")
        ax.legend(loc='best')
        if influence_vals:
            max_label, _ = self.get_max_k(influence_vals, 1)
            max_label = max_label[0]
            result_bar[labels.index(max_label)].set_color('green')

        ax.set_xticks(ind + width / 2)
        label_tags = tuple([utils.to_valid_latex(bin_item.get_bin_representation(i)) for i in labels])
        tags_max_length = max([len(tag) for tag in label_tags])
        ax.set_xticklabels(label_tags, rotation='vertical' if tags_max_length >= 4 else 'horizontal')

        ax.set_xlabel(utils.to_valid_latex(bin_item.get_bin_name() + " values"), fontsize=20)
        ax.set_ylabel("frequency(\\%)", fontsize=16)


        if title is not None:
            if show_scores:
                ax.set_title(f'score: {score}\n {utils.to_valid_latex(title)}', fontdict={'fontsize': 14})
            else:
                ax.set_title(utils.to_valid_latex(title), fontdict={'fontsize': 14})

        ax.set_axis_on()
        return bin_item.get_bin_name() ####

    def interestingness_only_explanation(self, source_col, result_col, col_name):
        if utils.is_categorical(source_col):
            vc = source_col.value_counts()
            source_max = utils.max_key(vc)
            vc = result_col.value_counts()
            result_max = utils.max_key(vc)
            return f"The distribution of column '{col_name}' changed significantly.\n" \
                   f"The most common value was {source_max} and now it is {result_max}."

        std_source = np.sqrt(np.var(source_col))
        mean_source = np.mean(source_col)
        std = np.sqrt(np.var(result_col))
        mean = np.mean(result_col)

        return f"The distribution of column '{col_name}' changed significantly.\n" \
               f" The mean was {mean_source:.2f} and the standard " \
               f"deviation was {std_source:.2f}, and now the mean is {mean:.2f} and the standard deviation is {std:.2f}."

    def calc_measure_internal(self, bin: Bin):
        return ExceptionalityMeasure.kstest(bin.source_column.dropna(),
                                            bin.result_column.dropna())  # / len(source_col.dropna().value_counts())

    def sat(self, d1, d2, attr, cond, k, cont, att, rem_attr=None, n=1):
        total = 0
        max_val = 0
        # if n == 1:
            # mem = self.mem1
        # else:
            # mem = self.mem2
        ret = 0

        for i in range(2, k):
            # if k in mem.keys():
                # total += mem[k]
                # if mem[k] > max_val:
                    # max_val = mem[k]
                # continue
            subsets_1 = list(combinations(d1.index, i))
            subset_dfs_1 = [d1.loc[list(subset)] for subset in subsets_1]

            subsets_2 = list(combinations(d2.index, k-i))
        # print(len(subsets_2))
            subset_dfs_2 = [d2.loc[list(subset)] for subset in subsets_2]
            for sdf1 in subset_dfs_1:
                # if str(sdf1) in self.mem_comb.keys():
                    # ret += self.mem_comb[str(sdf1)]
                    # continue
                num = 0
                sum = 0
                # print(total)
                for sdf2 in subset_dfs_2:
                    df_join = pd.merge(sdf1, sdf2, on=attr, how='inner')
                    if df_join.shape[0]:
                        
                    # print(f'--------------------------------\n{sdf1}\n**************\n{sdf2}')
                        # total += 1
                        val = df_join[att].agg(cont)
                        if val > max_val:
                            max_val = val
                        # mem[k] = val
                        sum += val
                        total += 1
                        num += 1
                ret += sum / (num)
                # ret = self.mem_comb[str(sdf1)]
                
        # if n == 1:
            # self.mem1 = mem 
        # else:
            # self.mem2 = mem 
        if total == 0:
            return 0
        return ret# sum / num# + np.log2(total)# / num)
    
    def shapley(self, d1, d1_exc, d2, attr, cond, sat,cont, att):
        # m = min(d1.shape[0] + d2.shape[0], 3*d1.shape[0])
        m = 5
        including = 0
        excluding = 0
        self.mem1 = {}
        self.mem2 = {}
        for k in range(m):
            if k not in self.including.keys():
                self.including[k] =  ((math.factorial(k)*math.factorial(m - k - 1))/math.factorial(m)) * sat(d1, d2, attr, cond, k, n=1, cont=cont, att=att)
            including += self.including[k]
            # including += ((math.factorial(k)*math.factorial(m - k - 1))/math.factorial(m)) * sat(d1, d2, attr, cond, k, n=1, cont=cont, att=att)
            excluding += ((math.factorial(k)*math.factorial(m - k - 1))/math.factorial(m)) * sat(d1_exc, d2, attr, cond, k, n=2, cont=cont, att=att)
        return including - excluding
    
    def get_most_contributing_fact(self, op, d1, d2, attr, cond, consider='right', top_k=1, cont=None, att=None):
        if att == None:
            cont = 'count'
            att = list(list(facts.values())[0].reset_index().to_dict().items())[0]
        self.consider = consider
        self.including = {}
        self.mem_comb = {}
        top_shapley = 0
        top_fact = None
        facts = {}
        if consider == 'right':
            d1_consider = d2
            d2_consider = d1
        else:
            d1_consider = d1
            d2_consider = d2
        d2_consider = d2_consider[d2_consider[attr].isin(pd.merge(d1_consider, d2_consider, on=attr, how='inner')[attr])]
        for i in (d1_consider.index):
            # print(f'{i}')#, {d1_consider.loc[i][attr]}, {pd.merge(d1_consider, d2_consider, on=attr, how='inner')[attr].values}')
            if d1_consider.loc[i][attr] not in pd.merge(d1_consider, d2_consider, on=attr, how='inner')[attr].values:
                v_shapley = 0
            else:
                d_tmp = d1_consider.drop(index=i, inplace=False)
                v_shapley = self.shapley(d1_consider, d_tmp, d2_consider, attr, cond, self.sat, cont, att)
            facts[str(d1_consider.loc[i])] = [v_shapley, d1_consider[d1_consider.index == i]]
            if v_shapley > top_shapley:
                top_shapley = v_shapley
                top_fact = d1_consider[d1_consider.index == i]
        pass
        facts = {k: v[1] for k, v in sorted(facts.items(), key=lambda item: -(item[1][0]))}
    # print(f'The top fact here is:\n{str(top_fact)}')
        fig, axes = plt.subplots(1, figsize=(14, 2))

        axes.axis('off')
        # axes[1].axis('off')
        d1_name = r'$\bf{{{}}}$'.format(utils.to_valid_latex(d1.df_name, True))
        d2_name = r'$\bf{{{}}}$'.format(utils.to_valid_latex(d2.df_name, True))
        if consider == 'left':
            consider = d1_name
        else:
            consider = d2_name
        bold_attr = r'$\bf{{{}}}$'.format(utils.to_valid_latex(attr), True)
        
        if top_k == 1:
            explanation = f'When joining the DataFrames\nThe most significant fact from DataFrame {consider} is:\n'
        else:
            explanation = f'When joining the DataFrames\nThe {top_k} most significant facts from DataFrame {consider} in descending order are:'
        # explanation = f'The tuple '
        fact_exp = ''
        n = 0
        while n < top_k:
            i = 0
            explanation += f'\n'
            for k, v in list(facts.values())[n].reset_index().to_dict().items():
                if i != 0:
                    explanation += ', '
                i += 1
            # v = r'$\bf{{{}}}$'.format(utils.to_valid_latex(v[0], True))
                # r'$\bf{{{}}}$'.format(utils.to_valid_latex(f'{k}={v[0]}', True))
                
                # explanation += f'{r'$\bf{{{}}}$'.format(utils.to_valid_latex(k, True))}={r'$\bf{{{}}}$'.format(utils.to_valid_latex(v[0], True))}'
            n += 1
        # explanation += f'\nfrom DataFrame {consider} is {r'$\bf{very}$'} {r'$\bf{significant}$'} for this result.'
        props1 = dict(boxstyle='round', facecolor='white', alpha=0.5)
        axes.text(0.5, 0.5, explanation, transform=axes.transAxes, fontsize=20,
        verticalalignment='center', horizontalalignment='center', bbox=props1)

        props2 = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
        # axes[1].text(0.5, 0.5, fact_exp, transform=axes[1].transAxes, fontsize=14,
        # verticalalignment='center', horizontalalignment='center', bbox=props2)
        fig.show()
        # txt = ax.text(x, y, n, fontsize=10)
        return top_fact, facts

    def calc_influence_col(self, current_bin: Bin):
        self.including = None
        bin_values = current_bin.get_bin_values()
        source_col = current_bin.get_source_by_values(bin_values)
        res_col = current_bin.get_result_by_values(bin_values)
        if len(bin_values) > 15:
            bin_values = list(source_col.value_counts().nlargest(15).keys())
        score_all = ExceptionalityMeasure.kstest(source_col, res_col)
        influence = []
        for value in bin_values:
            source_col_only_list = current_bin.get_source_by_values([b for b in bin_values if b != value])
            res_col_only_list = current_bin.get_result_by_values([b for b in bin_values if b != value])

            score_without_bin = ExceptionalityMeasure.kstest(source_col_only_list, res_col_only_list)
            influence.append(score_all - score_without_bin)

        return influence

    def build_operation_expression(self, source_name):
        from fedex_generator.Operations.Filter import Filter
        from fedex_generator.Operations.Join import Join

        if isinstance(self.operation_object, Filter):
            return f'Dataframe {self.operation_object.source_name}, ' \
                   f'filtered on attribute {self.operation_object.attribute}'
        elif isinstance(self.operation_object, Join):
            return f'{self.operation_object.right_name} joined with {self.operation_object.left_name} by {self.operation_object.attribute}'

    def build_explanation(self, current_bin: Bin, col_name, max_value, source_name):
        source_col = current_bin.get_binned_source_column()
        res_col = current_bin.get_binned_result_column()

        res_probs = res_col.value_counts(normalize=True)
        source_probs = source_col.value_counts(normalize=True)
        for bin_value in current_bin.get_bin_values():
            res_probs[bin_value] = res_probs.get(bin_value, 0)
            source_probs[bin_value] = source_probs.get(bin_value, 0)

        additional_explanation = []
        if current_bin.name == "NumericBin":
            values = current_bin.get_bin_values()
            index = values.index(max_value)

            values_range_str = "below {}".format(utils.format_bin_item(values[1])) if max_value == 0 else \
                "above {}".format(utils.format_bin_item(max_value)) if index == len(values) - 1 else \
                    "between {} and {}".format(utils.format_bin_item(values[index]),
                                               utils.format_bin_item(values[index + 1]))
            factor = res_probs.get(max_value, 0) / source_probs[max_value]   
            proportion = "less" if factor < 1 else "more"  
            if (factor < 1 and factor > 0):
                factor = 1 / factor
            if factor == 0:
                additional_explanation.append(
                    f"{START_BOLD}{utils.to_valid_latex(col_name, True)}{END_BOLD} values "
                    f"{START_BOLD}{utils.to_valid_latex(values_range_str, True)}{END_BOLD}\n"
                    f"are {START_BOLD}no{END_BOLD} {START_BOLD}longer{END_BOLD} {START_BOLD}exist{END_BOLD} (was {round(source_probs[max_value],3)*100}%)")
            else:
                appear_test = f'{utils.smart_round(factor)} times {proportion}'
                additional_explanation.append(
                    f"{START_BOLD}{utils.to_valid_latex(col_name, True)}{END_BOLD} values "
                    f"{START_BOLD}{utils.to_valid_latex(values_range_str, True)}{END_BOLD} (in green)\n"
                    f"appear {START_BOLD}"
                    f"{utils.to_valid_latex(appear_test, True)}"
                    f"{END_BOLD} than before")
        else:
            factor = res_probs.get(max_value, 0) / source_probs[max_value]

            source_prob = 100 * source_probs[max_value]
            res_prob = 100 * res_probs.get(max_value, 0)
            max_value_rep = current_bin.get_bin_representation(max_value)

            if factor == 0:
                proportion_sentes = f' frequency was {source_prob:.1f}% now {res_prob:.1f}%'
            elif factor < 1:
                proportion = "less"
                factor = 1 / factor
                proportion_sentes = f"appear {START_BOLD} " \
                                    f"{utils.to_valid_latex(f'{utils.smart_round(factor)} times {proportion}', True)}" \
                                    f" {END_BOLD} than before"
            else:
                proportion = "more"
                proportion_sentes = f"appear {START_BOLD} " \
                                    f"{utils.to_valid_latex(f'{utils.smart_round(factor)} times {proportion}', True)}" \
                                    f" {END_BOLD} than before"

            additional_explanation.append(
                f"{START_BOLD}{utils.to_valid_latex(col_name, True)}{END_BOLD} value "
                f"{START_BOLD}{utils.to_valid_latex(max_value_rep, True)}"
                f"{END_BOLD} (in green)\n{proportion_sentes}")

        influence_top_example = ", ".join(additional_explanation)
        
        return influence_top_example



import pandas as pd
from pandas import DataFrame

from fedex_generator.commons.consts import TOP_K_DEFAULT, DEFAULT_FIGS_IN_ROW
from fedex_generator.commons import utils
from fedex_generator.commons.DatasetRelation import DatasetRelation
from fedex_generator.Operations import Operation
from fedex_generator.Measures.DiversityMeasure import DiversityMeasure

from typing import Generator, List, Tuple


class GroupBy(Operation.Operation):
    """
    Implementation of the GroupBy operation, fit for the FEDEx explainability framework.\n
    Provides a .explain() method for explaining the operation, as well as methods used for producing the explanation.
    """

    def __init__(self, source_df, source_scheme, group_attributes, agg_dict, result_df=None, source_name=None,
                 operation=None, column_mapping=None):
        """
        :param source_df: The source DataFrame, before the groupby operation.
        :param source_scheme: The scheme of the source DataFrame.
        :param group_attributes: The attributes to group by.
        :param agg_dict: The aggregation dictionary.
        :param result_df: The resulting DataFrame after the groupby operation.
        :param source_name: The name of the source DataFrame.
        :param operation: The operation to perform.
        :param column_mapping: A possible mapping between the source and result columns, if changes were made.
        """
        super().__init__(source_scheme)
        # Set the attributes
        self.source_scheme = source_scheme
        self.group_attributes = group_attributes
        self.agg_dict = agg_dict
        self.source_name = source_name
        self.source_df = source_df
        self.source_name = utils.get_calling_params_name(source_df)

        # If the result DataFrame is None, perform the groupby and aggregation
        if result_df is None:
            self.source_name = utils.get_calling_params_name(source_df)
            source_df = source_df.reset_index()
            group_attributes = self.get_one_to_many_attributes(source_df, group_attributes)
            self.result_df = source_df.groupby(group_attributes).agg(agg_dict)
        else:
            self.result_df = result_df
            self.result_name = utils.get_calling_params_name(result_df)

        self._column_mapping = column_mapping
        self._measure = None

    def iterate_attributes(self) -> Generator[Tuple[str, DatasetRelation], None, None]:
        """
        Iterate over the attributes of the result DataFrame.

        This generator function yields each attribute of the result DataFrame along with a DatasetRelation object with the result DF.
        It skips the attribute if it is named 'index'.

        :yield: A tuple containing the attribute name and a DatasetRelation object.
        """
        for attr in self.result_df.columns:
            if isinstance(attr, str) and attr.lower() == "index":
                continue
            yield attr, DatasetRelation(None, self.result_df, self.source_name)

    def get_source_col(self, filter_attr, filter_values, bins):
        return None

    def explain(self, schema: dict = None, attributes: List[str] = None, top_k: int = TOP_K_DEFAULT,
                explainer: str = 'fedex',
                figs_in_row: int = DEFAULT_FIGS_IN_ROW, show_scores: bool = False, title: str = None,
                corr_TH: float = 0.7, consider='right', cont=None, attr=None, ignore=[],
                use_sampling=True, sample_size: int | float = Operation.SAMPLE_SIZE,
                debug_mode: bool = False, draw_figures: bool = True, return_scores: bool = False,
                measure_only: bool = False
                ) -> (None | Tuple[str, pd.Series, int, int, pd.Series, pd.Series, pd.Series, str, bool] | Tuple):
        """
        Explain for group by operation
        :param schema: dictionary with new columns names, in case {'col_name': 'i'} will be ignored in the explanation
        :param attributes: only this attributes will be included in the explanation calculation
        :param top_k: top k explanations number, default one explanation only.
        :param show_scores: show scores on explanation
        :param figs_in_row: number of explanations figs in one row
        :param title: explanation title
        :param dir: direction of the outlier. Can be 'high' or 'low', or the corresponding integer values 1 and -1 (HIGH and LOW constants).
        :param use_sampling: whether to use sampling for the explanation.
        :param sample_size: the sample size to use when sampling the data. Can be an integer or a float between 0 and 1. Default is 5000.
        :param debug_mode: Developer option. Disables multiprocessing for easier debugging. Default is False. Can possibly add more debug options in the future.
        :param draw_figures: Whether or not to draw the figures in this stage. Defaults to True. Set to False if you want to draw the figures later.

        :return: explain figures
        """
        if schema is None:
            schema = {}

        if attributes is None:
            attributes = []

        backup_source_df, backup_res_df = None, None
        if use_sampling:
            backup_source_df, backup_res_df = self.source_df, self.result_df
            self.source_df, self.result_df = self.sample(self.source_df, sample_size), self.sample(self.result_df,
                                                                                                   sample_size)

        # Unless the outlier explainer is used, the diversity measure is always used for the groupby operation.
        measure = DiversityMeasure()
        self._measure = measure
        scores = measure.calc_measure(self, schema, attributes, ignore=ignore, unsampled_source_df=backup_source_df,
                                      unsampled_res_df=backup_res_df, column_mapping=self._column_mapping,
                                      debug_mode=debug_mode)

        if use_sampling:
            self.source_df, self.result_df = backup_source_df, backup_res_df

        if measure_only:
            return scores

        ret_val = measure.calc_influence(utils.max_key(scores), top_k=top_k, figs_in_row=figs_in_row,
                                          show_scores=show_scores, title=title, debug_mode=debug_mode,
                                          draw_figures=draw_figures)
        if return_scores:
            return ret_val, scores
        else:
            return ret_val, None


    def draw_figures(self, title: str, scores: pd.Series, K: int, figs_in_row: int, explanations: pd.Series, bins: pd.Series,
                      influence_vals: pd.Series, source_name: str, show_scores: bool,
                     added_text: dict | None = None, added_text_name: str | None = None) -> tuple:
        return self._measure.draw_figures(
            title=title, scores=scores, K=K, figs_in_row=figs_in_row,
            explanations=explanations, bins=bins, influence_vals=influence_vals,
            source_name=source_name, show_scores=show_scores, added_text=added_text,
            added_text_name=added_text_name
        )

    @staticmethod
    def get_one_to_many_attributes(df, group_attributes):
        """
        Identify and append one-to-many relationship attributes to the group attributes list.

        This function iterates over the columns of the DataFrame and checks if there is a one-to-many relationship
        between the specified group attributes and other columns. If such a relationship is found, the corresponding
        column is appended to the group attributes list.

        :param df: The DataFrame to check for one-to-many relationships.
        :param group_attributes: The list of attributes to group by.
        :return: The updated list of group attributes including one-to-many relationship attributes.
        """
        for col in group_attributes:
            for candidate_col in df:
                if candidate_col in group_attributes:
                    continue

                if GroupBy._is_one_to_many(df, col, candidate_col):
                    group_attributes.append(candidate_col)

        return group_attributes

    @staticmethod
    def _is_one_to_many(df: DataFrame, col1: str, col2: str) -> bool:
        """
        Check if there is a one-to-many relationship between two columns in a DataFrame.

        This function determines if there is a one-to-many relationship between the specified columns
        by checking if each unique value in `col1` maps to a unique value in `col2`.

        :param df: The DataFrame to check for the one-to-many relationship.
        :param col1: The first column to check.
        :param col2: The second column to check.
        :return: True if there is a one-to-many relationship, False otherwise.
        """
        if type(df) != pd.DataFrame:
            df = pd.DataFrame(df)

        # Check if the first 1000 rows have a one-to-many relationship. This is done first for performance reasons.
        first_max_cheap_check = df[[col1, col2]].head(1000).groupby(col1).nunique()[col2].max()
        # If the first 1000 rows do not have a one-to-many relationship, return False
        if first_max_cheap_check != 1:
            return False

        # Check if the entire DataFrame has a one-to-many relationship
        first_max = df[[col1, col2]].groupby(col1).nunique()[col2].max()
        return first_max == 1

    def _get_columns_names(self):
        """
        Generate a list of column names for the result DataFrame.

        This method processes the columns of the result DataFrame and constructs a list of column names.
        It handles cases where columns are tuples (e.g., from multi-index DataFrames) by joining the tuple elements.
        It also appends the aggregation method to the column names if applicable.

        :return: A list of column names for the result DataFrame.
        """
        columns = []
        for column in list(self.result_df.columns):
            if isinstance(column, tuple):
                columns.append("_".join(column))
            else:
                if column in self.agg_dict:
                    columns.append(f'{column}_{self.agg_dict[column]}')
                elif isinstance(self.agg_dict, str) and not column.endswith(f'_{self.agg_dict}'):
                    columns.append(f'{column}_{self.agg_dict}')
                else:
                    columns.append(f'{column}')

        return columns



# Constants used in the project

# Significance threshold defines a threshold for the significance of the feature for it to be considered at all
SIGNIFICANCE_THRESHOLD = 0.1
# Default number of top explanations to show
TOP_K_DEFAULT = 1
# Default number of figures in one row
DEFAULT_FIGS_IN_ROW = 2



import math
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Dict, List, Tuple, Any
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import rc
import numpy as np
import pandas as pd
from matplotlib.figure import Figure
from paretoset import paretoset

from fedex_generator.Measures.Bins import Bins, Bin
from fedex_generator.commons.consts import SIGNIFICANCE_THRESHOLD, TOP_K_DEFAULT, DEFAULT_FIGS_IN_ROW
from fedex_generator.commons.utils import is_numeric, to_valid_latex

try:
    from FEDEx_Generator.src.fedex_generator.Operations.Operation import Operation
    from FEDEx_Generator.src.fedex_generator.commons.DatasetRelation import DatasetRelation
except:
    from fedex_generator.Operations.Operation import Operation
    from fedex_generator.commons.DatasetRelation import DatasetRelation

# Sets some parameters for the plots. Usage of latex is optional, and is currently set to False by default.
usetex = False  # matplotlib.checkdep_usetex(True)
print(f"usetex-{usetex}")
rc('text', usetex=usetex)
matplotlib.rcParams.update({'font.size': 16})

# Bold text in latex, for the plots
START_BOLD = r'$\bf{'
END_BOLD = '}$'


def draw_pie(items_dict: dict, important_item=None) -> None:
    """
    Draw a pie chart of the items_dict. The important_item will be highlighted.
    :param items_dict: dictionary of items and their probabilities
    :param important_item: the item to highlight
    """
    labels = items_dict.keys()
    probabilities = [items_dict[item] for item in labels]
    explode = [0.1 if item == important_item else 0 for item in labels]

    fig1, ax1 = plt.subplots()
    ax1.pie(probabilities, explode=explode, labels=labels, autopct='%1.1f%%', shadow=True, startangle=90)
    ax1.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.

    plt.show()


def draw_histogram(items_before: list, items_after: list, label: str, title) -> None:
    """
    Draw a histogram of the items_before and items_after.
    :param items_before: list of items before the operation
    :param items_after: list of items after the operation
    :param label: the label of the histogram
    :param title: the title of the histogram
    """
    if items_before is None:
        items_before = []
    all_vals = list(set(list(items_before) + list(items_after)))
    # If the values are not numeric, we can't use bins, and we just plot the values.
    if not is_numeric(all_vals):
        if len(items_before) > 0:
            plt.hist([sorted(list(items_before)), sorted(list(items_after))], label=["Before", "After"],
                     color=[u'#ff7f0e', u'#1f77b4'])
        else:
            plt.hist([sorted(list(items_after))], label=["After"])

    # If the values are numeric, we use bins.
    else:
        bins = np.linspace(np.min(all_vals), np.max(all_vals), min(40, len(all_vals)))
        if len(items_before) > 0:
            plt.hist([list(items_before), list(items_after)], bins=bins, label=["Before", "After"],
                     color=[u'#ff7f0e', u'#1f77b4'])
        else:
            plt.hist([list(items_after)], bins=bins, label=["After"])

    plt.ylabel('results occurrences count', fontsize=16)
    plt.xlabel(to_valid_latex(label), fontsize=16)
    plt.legend(loc='upper right')
    plt.title(to_valid_latex(title))
    plt.show()


class BaseMeasure(object):
    """
    Base class for all measures. Contains the basic methods for calculating the measure and building the explanation.
    Inherit from this class to create a new measure, and implement the abstract methods.\n
    """

    def __init__(self):
        self.source_dict, self.max_val, self.score_dict, self.operation_object, self.scheme = \
            None, None, None, None, None
        self.bins = {}
        self.bin_func = pd.qcut

    def interestingness_only_explanation(self, source_col: pd.Series, result_col: pd.Series, col_name: str) -> str:
        """
        Provide an explanation, based only on the interestingness score.\n
        Abstract method, should be implemented by the inheriting class.
        :param source_col: The column from the source DataFrame
        :param result_col: The column from the result DataFrame
        :param col_name: The name of the column
        :return: The explanation string
        """
        raise NotImplementedError()

    @staticmethod
    def get_source_and_res_cols(dataset_relation: DatasetRelation, attr: str) -> Tuple[pd.Series, pd.Series]:
        """
        Static method to get source and result columns from dataset_relation.
        :param dataset_relation: dataset relation object
        :param attr: attribute name
        :return source_col, res_col: source and result columns with the attribute name, without null values
        """
        source_col, res_col = dataset_relation.get_source(attr), dataset_relation.get_result(attr)
        res_col = res_col[~res_col.isnull()]
        if source_col is not None:
            source_col = source_col[~source_col.isnull()]

        return source_col, res_col


    def _calc_measure(self, attr: str, dataset_relation: DatasetRelation, operation_object: Operation, scheme: dict,
                      use_only_columns: list, ignore=None,
                      unsampled_source_df: pd.DataFrame = None, unsampled_res_df: pd.DataFrame = None,
                      column_mapping: dict=None) \
            -> Tuple[str, float, str, Bins, Tuple[pd.Series, pd.Series]] | Tuple[str, float, None, None, None]:
        """

        :param attr: The attribute name.
        :param dataset_relation: The dataset relation for the current attribute.
        :param operation_object: The operation object for the current operation.
        :param scheme: The scheme of the columns.
        :param use_only_columns: A list of columns to include in the explanation. If empty, all columns will be included.
        :param ignore: A list of columns to ignore in the explanation. Set to [] by default.
        :param unsampled_source_df: The source DataFrame before sampling. Optional. Only needed if sampling is used.
        :param unsampled_res_df: The result DataFrame before sampling. Optional. Only needed if sampling is used.
        :param column_mapping: A dict mapping the original column names to the current column names. Optional. Needed in case some columns were renamed as part of a groupby operation.
        :return:
        """
        # If the attribute is in the ignore list, skip it.
        if attr in ignore:
            return attr, -1, None, None, None

        # Get the column scheme from the scheme dictionary. If not found, set it to 'ni'.
        column_scheme = scheme.get(attr, "ni").lower()
        # If the column scheme is 'i', skip the attribute.
        if column_scheme == "i":
            return attr, -1, None, None, None

        # If there are columns in the use_only_columns list, and the attribute is not in the list, skip it.
        if len(use_only_columns) > 0 and attr not in use_only_columns:
            return attr, -1, None, None, None

        # Get the source and result columns for the attribute.
        source_col, res_col = self.get_source_and_res_cols(dataset_relation, attr)
        # If the result column is empty, skip the attribute.
        if len(res_col) == 0:
            return attr, -1, None, None, None


        # Create bin candidates from the source and result columns, with a bin count specified by the operation object.
        size = operation_object.get_bins_count()

        bin_candidates = Bins(source_col, res_col, size)
        unsampled_bin_candidates = None

        if unsampled_source_df is not None and unsampled_res_df is not None:
            # source_col can be none in some groupby cases. We don't want to modify it if it is, as that may cause
            # unexpected behavior.
            if source_col is not None:
                # In the case of groupby that creates tuple columns, we need to handle the case where the attribute in the
                # result dataframe is a tuple that did not exist in the source dataframe.
                if attr not in unsampled_source_df.columns and isinstance(attr, tuple):
                    source_col = unsampled_source_df[attr[0]]
                else:
                    source_col = unsampled_source_df[attr] if attr not in column_mapping else unsampled_source_df[column_mapping[attr]]
            res_col = unsampled_res_df[attr]
            unsampled_bin_candidates = Bins(source_col, res_col, size)

        # Compute the measure score for each bin candidate, and get the maximum score.
        measure_score = -np.inf
        for bin_ in bin_candidates.bins:
            measure_score = max(self.calc_measure_internal(bin_), measure_score)


        return (attr, measure_score, dataset_relation.get_source_name(),
                unsampled_bin_candidates if unsampled_bin_candidates is not None else bin_candidates,
                (source_col, res_col))

    def calc_measure(self, operation_object: Operation, scheme: dict, use_only_columns: list, ignore=None,
                     unsampled_source_df: pd.DataFrame = None, unsampled_res_df: pd.DataFrame = None,
                     column_mapping: dict=None, debug_mode: bool = False
                     ) -> \
            Dict[str, float]:
        """
        Calculate the measure for each attribute in the operation_object.

        :param operation_object: The operation object - the operation that was performed, one of the classes in the Operations package.
        :param scheme: The scheme of the columns, in the form of a dictionary, where the key is the column name and the value is the scheme of the column. Columns with the scheme 'i' will be ignored.
        :param use_only_columns: A list of columns to include in the explanation. If empty, all columns will be included.
        :param ignore: A list of columns to ignore in the explanation. Set to [] by default.
        :param unsampled_source_df: The unsampled source DataFrame. Optional. Only needed if sampling is used.
        :param unsampled_res_df: The unsampled result DataFrame. Optional. Only needed if sampling is used.
        :param column_mapping: A dict mapping the original column names to the current column names. Optional. Needed in case some columns were renamed as part of a groupby operation.
        :param debug_mode: Developer option. Disables multiprocessing for easier debugging. Default is False.
        """
        # Set the operation object, the scheme, and the score dictionary to the given values.
        # Also initialize the max_val to -1.
        if ignore is None:
            ignore = []
        if column_mapping is None:
            column_mapping = {}

        self.operation_object = operation_object
        self.score_dict = {}
        self.max_val = -1
        self.scheme = scheme

        # Iterate over the attributes in the operation object.
        # From limited testing, doing this in parallel gives a small performance boost.
        if not debug_mode:
            with ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(self._calc_measure, attr, dataset_relation, operation_object, scheme, use_only_columns,
                                    ignore, unsampled_source_df, unsampled_res_df, column_mapping) for attr, dataset_relation in operation_object.iterate_attributes()
                ]
                for future in as_completed(futures):
                    attr, measure_score, source_name, bins, cols = future.result()
                    if measure_score != -1:
                        self.score_dict[attr] = (source_name, bins, measure_score, cols)
        else:
            for attr, dataset_relation in operation_object.iterate_attributes():
                attr, measure_score, source_name, bins, cols = self._calc_measure(attr, dataset_relation, operation_object, scheme, use_only_columns, ignore, unsampled_source_df, unsampled_res_df, column_mapping)
                if measure_score != -1:
                    self.score_dict[attr] = (source_name, bins, measure_score, cols)

        # Get the maximum value from the score dictionary and set it to the max_val attribute.
        self.max_val = max([kl_val for _, _, kl_val, _ in self.score_dict.values()])

        # Return a dictionary with the attribute names as keys and the measure scores as values.
        return dict([(attribute, _tuple[2]) for (attribute, _tuple) in self.score_dict.items()])

    def calc_measure_internal(self, _bin: Bin) -> float:
        """
        Calculate the measure score for a bin.
        This is an abstract method, and should be implemented by the inheriting class for the specific measure.
        :param _bin: The bin object.
        :return: The measure score.
        """
        raise NotImplementedError()

    def build_operation_expression(self, source_name: str) -> str:
        """
        Get manipulation expression

        :param source_name: The source df name
        :return: string of the manipulation
        """
        raise NotImplementedError()

    def build_explanation(self, current_bin: Bin, max_col_name: str, max_value: float, source_name: str) -> str:
        """
        Build an explanation for the given bin, column name, value, and source name after calculating the measure score.
        This is an abstract method, and should be implemented by the inheriting class.
        :param current_bin: The current bin object.
        :param max_col_name: The column name with the maximum measure score.
        :param max_value: The maximum value of the column.
        :param source_name: The source name.
        """
        raise NotImplementedError()

    def get_influence_col(self, current_bin: Bin) -> Dict[Any, float]:
        """
        Compute and return the influence of the column in the current bin.
        :param current_bin: The current bin object.
        :return: A dictionary with the bin values as keys and the influence values as values.
        """
        bin_values = current_bin.get_bin_values()
        return dict(zip(bin_values, self.calc_influence_col(current_bin)))

    def calc_influence_col(self, current_bin: Bin) -> List[float]:
        """
        Calculate the influence of the column in the current bin.
        This is an abstract method, and should be implemented by the inheriting class.\n
        For FEDEX, the influence of a set of rows is defined as:\n
        .. math:: I_A (d_{in}, q, d_{out}) - I_A (d_{in} - R, q, d_{out}')\n
        Where:\n
        - :math:`I_A (d_{in}, q, d_{out})` is the measure score of the column A in the current bin.
        - :math:`d_{in}` is the input dataframe.
        - :math:`d_{out}` is the output dataframe.
        - :math:`q` is the query.
        - :math:`R` is the set of rows in the current bin.
        - :math:`d_{out}'` is the output dataframe after removing the rows in the current bin.
        :param current_bin: The current bin object.
        :return: A list of influence values. Each value corresponds to a value in the bin.
        """
        raise NotImplementedError()

    def get_max_k(self, score_dict, k) -> Tuple[List[str], List[float]]:
        """
        Get the top k attributes with the highest scores from the score dictionary.

        This method sorts the attributes based on their scores and returns the top k attributes along with their scores.

        :param score_dict: A dictionary where keys are attribute names and values are their corresponding scores.
        :param k: The number of top attributes to return.
        :return: A tuple containing two lists - the top k attribute indexes and their corresponding scores.
        """
        # Get the keys and values from the score dictionary, convert them to arrays,
        # and get the sorted indices of the values.
        score_array_keys = list(score_dict.keys())
        score_array = np.array([score_dict[i] for i in score_array_keys])
        max_indices = score_array.argsort()

        unique_max_indexes = []
        influence_vals_max_indexes = []

        # Iterate over the indices, and add the values and keys to the above 2 lists, in ascending order.
        for index in max_indices:
            current_index = score_array_keys[index]
            unique_max_indexes.append(current_index)
            influence_vals_max_indexes.append(score_dict[current_index])

        # Get the top k attributes and their scores, then return them.
        max_indices = unique_max_indexes[-k:][::-1]
        max_influences = influence_vals_max_indexes[-k:][::-1]
        return max_indices, max_influences



    def _try_fix_explanation(self, explanation: str, binned_column: pd.Series, max_value) -> str:
        """
        If the explanation is not valid, for example in the case of group-bys where we have a value instead
        of the group names, try to fix it.
        Override this method in the inheriting class if needed, otherwise leave as is (returns the explanation as is).
        :param explanation: The explanation to fix.
        :param binned_column: The binned column.
        :param max_value: The max value of the bin.
        :param max_group_value: The group with the max value.
        :return:
        """
        return explanation

    def draw_bar(self, bin_item: Bin, influence_vals: dict = None, title: str = None, ax=None, score=None,
                 show_scores: bool = False, explanation_num: int | None = None) -> None:
        """
        Draw a bar plot for the given bin item and influence values.
        This is an abstract method, and should be implemented by the inheriting class.
        :param bin_item: The bin item to draw the bar plot for.
        :param influence_vals: The influence values for the bin item.
        :param title: The title of the plot. Optional.
        :param ax: The axes to draw the plot on. Optional.
        :param score: The score of the bin item. Optional.
        :param show_scores: Whether to show the scores on the plot. Optional.
        :param explanation_num: The explanation number to display in the title. Optional.
        """
        raise NotImplementedError()

    @staticmethod
    def get_significance(influence, influence_vals_list) -> float:
        """
        A static method to calculate the significance of the influence value,
        in the context of a list of influence values.\n
        Computed as the difference between the influence value and the mean of the influence values,
        divided by the square root of the variance of the influence values.
        :param influence: The influence value.
        :param influence_vals_list: The list of influence values.
        :return: The significance value of the influence.
        """
        influence_var = np.var(influence_vals_list)
        if influence_var == 0:
            return 0.0
        influence_mean = np.mean(influence_vals_list)

        return (influence - influence_mean) / np.sqrt(influence_var)


    def _calc_influence(self, score_dict, max_col_name, results_columns) -> pd.DataFrame:
        # This silly initialization is done because of the way pandas works with concatenation.
        # Concatenating an empty dataframe with another dataframe is deprecated, so we initialize it with a single empty row.
        # This row will be dropped later.
        results = pd.DataFrame([["", "", "", "", "", ""]], columns=results_columns)
        source_name, bins, score, _ = score_dict[max_col_name]

        for current_bin in bins.bins:
            # Compute the influence values for the current bin.
            influence_vals = self.get_influence_col(current_bin)
            influence_vals_list = np.array(list(influence_vals.values()))

            # If all the influence values are NaN, skip the current bin.
            if np.all(np.isnan(influence_vals_list)):
                continue

            # Get the top k (1 in this case) attribute indexes and their corresponding influence values.
            max_values, max_influences = self.get_max_k(influence_vals, 1)

            # Iterate over the top k attributes and their influence values.
            # Compute the significance of the influence value, and if it is above the threshold, build an explanation.
            for max_value, influence_val in zip(max_values, max_influences):
                significance = self.get_significance(influence_val, influence_vals_list)
                if significance < SIGNIFICANCE_THRESHOLD:
                    continue
                explanation = self.build_explanation(current_bin, max_col_name, max_value, source_name)
                explanation = self._try_fix_explanation(explanation, current_bin, max_value)

                # Save the results in a dictionary and append it to the results dataframe.
                new_result = dict(zip(results_columns,
                                      [score, significance, influence_val, explanation, current_bin, influence_vals,
                                       current_bin.get_bin_name(), max_col_name]))
                results = pd.concat([results, pd.DataFrame([new_result])], ignore_index=True)

        results = results.drop(axis='index', index=0)
        return results


    def draw_figures(self, title: str, scores: pd.Series, K: int, figs_in_row: int, explanations: pd.Series, bins: pd.Series,
                      influence_vals: pd.Series, source_name: str, show_scores: bool,
                     added_text: pd.Series = None, added_text_name: str | None = None)\
            -> tuple[list[str], plt.Figure] | None:
        """
        Draws the figures for the explanations.
        :param title: The title of the plot.
        :param scores: The scores of the attributes.
        :param K: The number of top attributes to consider.
        :param figs_in_row: The number of figures to display in a row.
        :param explanations: The explanations generated for the top K attributes.
        :param bins: The bins of the attributes.
        :param influence_vals: The influence values of the attributes.
        :param source_name: The name of the source DataFrame.
        :param show_scores: Whether to show the scores on the plot.
        :param added_text: Additional text to add to the bottom of each figure. Optional. A dict with explanation as key, and a sub-dict with 'text' and 'position' as keys.
        :param added_text_name: The name of the added text, to be displayed right above the added text.
        :return: A list of matplotlib figures containing the explanations for the top k attributes, after computing the influence.
        """
        figures = []
        # Set the title of the plot to the title if it is not None, otherwise build the operation expression.
        title = title if title else self.build_operation_expression(source_name)

        num_explanations = len(explanations) if explanations is not None else 0

        # If K is greater than 1, and there actually is more than one explanation, we need to set up a grid of subplots.
        # set the number of rows in the plot to the ceiling of the length of the scores divided by figs_in_row.
        if K > 1 and num_explanations > 1:  ###
            rows = math.ceil(len(scores) / figs_in_row)
            fig, axes = plt.subplots(rows, figs_in_row, figsize=(8 * figs_in_row, 9 * rows))
            for ax in axes.reshape(-1):
                ax.set_axis_off()
        else:
            total_text_len = 0
            if title:
                total_text_len += len(title)
            if explanations is not None and num_explanations > 0:
                total_text_len += len(explanations.iloc[0])
            # If the text is so long that it probably won't fit properly in the figure, increase the figure size.
            # Note that 300 is a fairly arbitrary threshold, made on an educated guess that the usual is around
            # 150-250 characters long, and that 300+ is probably around where it starts to get too long.
            if total_text_len > 300:
                fig, axes = plt.subplots(figsize=(9, 11))
            # Otherwise, use a smaller figure size.
            else:
                fig, axes = plt.subplots(figsize=(7, 8))

        # There are issues of the title not being properly centered when there is only one explanation,
        # so we need to check if there is only one explanation, and if so, set the title to be centered.
        if num_explanations == 1:
            # Why does x=0.6 work? I don't know, but it does and it centers the title properly.
            # Not 0.5, for some reason beyond my understanding.
            fig.suptitle(title, fontsize=15, y=1.02, x=0.6)
            axes.set_title(explanations.iloc[0], fontsize=16)
            axes.set_axis_off()
        else:
            fig.suptitle(title, fontsize=20, y=1.02)

        if num_explanations > 1:
            axes = axes.flatten()  # Flatten the axes array for easier indexing.
        else:
            axes = [axes]

        added_text_text = f"{added_text_name}:\n\n" if added_text_name else ""

        # Draw the bar plots for each explanation
        for index, (explanation, current_bin, current_influence_vals, score) in enumerate(
                zip(explanations, bins, influence_vals, scores)):

            explanation_num = index + 1 if (num_explanations > 1 and added_text is not None) else None

            figure = self.draw_bar(current_bin, current_influence_vals, title=explanation,
                                ax=axes[index], score=score,
                                show_scores=show_scores,
                                explanation_num=explanation_num
                                   )
            if figure:
                figures.append(figure)

            if added_text is not None:
                explanation_added_text = added_text.get(explanation, None)
                if explanation_added_text is not None:
                    explanation_added_text = explanation_added_text.get('added_text', "")
                    if not isinstance(explanation_added_text, str):
                        explanation_added_text = str(explanation_added_text)
                    # Replace any wrapping done previously - we want our own custom wrapping here, since we know
                    # what length of text we want to display.
                    explanation_added_text = explanation_added_text.replace("\n", " ")
                    if explanation_num is not None:
                        added_text_text += f"{START_BOLD}[{explanation_num}]{END_BOLD} {explanation_added_text}\n\n"
                    else:
                        added_text_text += f"{explanation_added_text}\n\n"


        # If there is text to add, add it to the bottom left of the plot.
        if added_text_text:
            plt.figtext(0, 0, added_text_text, horizontalalignment='left', verticalalignment='top',
                        fontsize=16, wrap=True,)

        # Adding a bit of a top margin to the plot, to make sure the title doesn't interfere with the plots.
        plt.subplots_adjust(top=0.92)
        plt.tight_layout()

        return figures, fig

    def calc_influence(self, brute_force=False, top_k=TOP_K_DEFAULT,
                       figs_in_row: int = DEFAULT_FIGS_IN_ROW, show_scores: bool = False, title: str = None,
                       deleted=None, debug_mode: bool = False, draw_figures: bool = True
                       ) -> Tuple[str, pd.Series, int, int, pd.Series, pd.Series, pd.Series, str, bool] | None | Figure | list[
        Figure] | list[str]:
        """
        Calculate the influence of each attribute in the dataset.

        This method computes the influence of each attribute by determining the change in the measure score
        when each attribute is removed from the dataset. The measure score is calculated using the `calc_measure` method.

        :param brute_force: Whether to use brute force to calculate the influence. Default is False.
        :param top_k: The number of top attributes to consider. Default is TOP_K_DEFAULT.
        :param figs_in_row: The number of figures to display in a row. Default is DEFAULT_FIGS_IN_ROW.
        :param show_scores: Whether to show the scores on the plot. Default is False.
        :param title: The title of the plot. Optional.
        :param deleted: A dictionary of deleted attributes as keys, with the values as a tuple: (dataframe name, bin object, score, column values). Optional.
        :param debug_mode: Developer option. Disables multiprocessing for easier debugging. Default is False.
        :param draw_figures: Whether to draw the figures at this stage. Default is True. If set to false, it is the responsibility of the caller to call the draw_figures method. Possible use case for leaving this false is if the caller wants to add additional information to the figures before drawing them.

        :return: A list (or a single) matplotlib figures containing the explanations for the top k attributes, after
        computing the influence. Alternatively, returns the names of the explained attributes, or None if no explanations were found.
        If the option draw_figures is set to False, returns a tuple containing the parameters needed to draw the figures instead.
        """

        # If deleted is not None, set the score dictionary to the deleted dictionary.
        # Else, set the score dictionary to the score dictionary of the measure.
        if deleted:
            score_dict = deleted
        else:
            score_dict = self.score_dict

        # Get a list of tuples of: (score, column name, bin object, column values) from the score_dict,
        # then sort that list
        score_and_col = [(score_dict[col][2], col, score_dict[col][1], score_dict[col][3])
                         for col in score_dict]

        list_scores_sorted = score_and_col
        list_scores_sorted.sort()

        # Define K, the number of top attributes to consider, as the value of top_k if deleted is None,
        # otherwise set it to the length of the deleted dictionary.
        K = top_k if not deleted else len(deleted.keys())

        # Create a dataframe for the results
        results_columns = ["score", "significance", "influence", "explanation", "bin", "influence_vals"]
        # This silly initialization is done because of the way pandas works with concatenation.
        # Concatenating an empty dataframe with another dataframe is deprecated, so we initialize it with a single empty row.
        # This row will be dropped later.
        results = pd.DataFrame([["", "","","","",""]], columns=results_columns)

        # Iterate over the top K attributes, and get the influence values for each bin.
        # From limited testing, doing this in parallel gives a small performance boost.
        if not debug_mode:
            with ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(self._calc_influence, score_dict, max_col_name, results_columns)
                    for score, max_col_name, bins, _ in list_scores_sorted[:-K - 1:-1]
                ]
                for future in as_completed(futures):
                    results = pd.concat([results, future.result()], ignore_index=True)
        else:
            for score, max_col_name, bins, _ in list_scores_sorted[:-K - 1:-1]:
                results = pd.concat([results, self._calc_influence(score_dict, max_col_name, results_columns)], ignore_index=True)

        results = results.drop(axis='index', index=0)

        # Compute the skyline of the results dataframe.
        results_skyline = results[results_columns[0:2]].astype("float")
        skyline = paretoset(results_skyline, ["diff", "max"])

        # Get the results for the skyline attributes.
        explanations = results[skyline]["explanation"]
        bins = results[skyline]["bin"]
        influence_vals = results[skyline]["influence_vals"]
        scores = results[skyline]["score"]

        source_name = score_dict[list_scores_sorted[-1][1]][0]

        # If there are no interesting explanations, print a message and return.
        if len(scores) == 0:
            print(f'Could not find any interesting explanations for your query over dataset {source_name}.')
            return None

        if draw_figures:
            figures = self.draw_figures(
                title=title,
                scores=scores,
                K=K,
                figs_in_row=figs_in_row,
                explanations=explanations,
                bins=bins,
                influence_vals=influence_vals,
                source_name=source_name,
                show_scores=show_scores,
            )
        else:
            return title, scores, K, figs_in_row, explanations, bins, influence_vals, source_name, show_scores

        # Return the figures if there are multiple figures, otherwise return a single figure.
        return figures

    def calc_interestingness_only(self) -> None:
        """
        Calculate and display a histogram based on the interestingness score.

        This method sorts the attributes based on their measure scores, selects the attribute with the highest score,
        and generates an explanation based solely on the interestingness score. It then draws a histogram to visualize
        the distribution of the source and result columns for the selected attribute.

        :return: None
        """
        # Get a list of tuples of: (score, column name, bin object, column values) from the score_dict,
        # then sort that list
        score_and_col = [(self.score_dict[col][2], col, self.score_dict[col][1], self.score_dict[col][3])
                         for col in self.score_dict]
        list_scores_sorted = score_and_col
        list_scores_sorted.sort()

        # Select the column name, bins and columns for the attribute with the highest score.
        _, col_name, bins, cols = list_scores_sorted[-1]

        # Get an explanation based solely on the interesting, then draw a histogram using it.
        io_explanation = self.interestingness_only_explanation(cols[0],
                                                               cols[1],
                                                               col_name)

        draw_histogram(cols[0], cols[1], col_name, io_explanation)



import pandas as pd
from pandas import DataFrame
from typing import Generator, Tuple, List

from fedex_generator.Measures.ShapleyMeasure import ShapleyMeasure
from fedex_generator.commons.consts import TOP_K_DEFAULT, DEFAULT_FIGS_IN_ROW
from fedex_generator.commons import utils
from fedex_generator.commons.DatasetRelation import DatasetRelation
from fedex_generator.Operations import Operation
from fedex_generator.Measures.ExceptionalityMeasure import ExceptionalityMeasure


class Join(Operation.Operation):
    """
    Implementation of the Join operation, fit for the FEDEx explainability framework.\n
    Provides a .explain() method for explaining the operation, as well as methods used for producing the explanation.
    """

    def __init__(self, left_df: DataFrame, right_df: DataFrame, source_scheme: dict, attribute: str,
                 result_df: DataFrame = None, left_name: str = None, right_name: str = None):
        """
        :param left_df: DataFrame to join on the left side.
        :param right_df: DataFrame to join on the right side.
        :param source_scheme: The scheme of the source DataFrame, as a dictionary.
        :param attribute: The attribute to join on.
        :param result_df: The resulting DataFrame after the join operation. Optional.
        :param left_name: Name of the left DataFrame. Optional.
        :param right_name: Name of the right DataFrame. Optional.
        """
        super().__init__(source_scheme)
        self.source_scheme = source_scheme
        self.attribute = attribute

        # If a result DataFrame is not provided, perform the join operation
        if result_df is None:
            left_name = utils.get_calling_params_name(left_df)
            right_name = utils.get_calling_params_name(right_df)
            left_df = left_df.copy().reset_index()
            right_df = right_df.copy()
            left_df.columns = [col if col in ["index", attribute] else left_name + "_" + col
                               for col in left_df]
            right_df.columns = [col if col in ["index", attribute] else right_name + "_" + col
                                for col in right_df]
            result_df = pd.merge(left_df, right_df, on=[attribute])

        self.right_name = right_name
        self.left_name = left_name
        self.left_df = left_df
        self.right_df = right_df
        self.result_df = result_df
        self._measure = None

    def iterate_attributes(self) -> Generator[Tuple[str, DatasetRelation], None, None]:
        """
        Iterate over the attributes of the left and right DataFrames.

        This method generates tuples containing an attribute and its corresponding DatasetRelation.
        It first yields attributes from the left DataFrame, then from the right DataFrame, excluding any attributes
        that are common between the two DataFrames.

        :yield: Tuples of attribute name and DatasetRelation objects with the left and right DataFrames and the result DataFrame.
        """
        for attr in self.left_df.columns:
            if isinstance(attr, str) and attr.lower() == "index":
                continue
            yield attr, DatasetRelation(self.left_df, self.result_df, self.left_name)

        for attr in set(self.right_df.columns) - set(self.left_df.columns):
            if isinstance(attr, str) and attr.lower() == "index":
                continue
            yield attr, DatasetRelation(self.right_df, self.result_df, self.right_name)

    def explain(self, schema: dict = None, attributes: List[str] = None, top_k: int = TOP_K_DEFAULT,
                figs_in_row: int = DEFAULT_FIGS_IN_ROW, show_scores: bool = False, title: str = None,
                corr_TH: float = 0.7, explainer='fedex', consider='right', cont=None, attr=None, ignore=[],
                use_sampling: bool = True, sample_size: int | float = Operation.SAMPLE_SIZE,
                debug_mode: bool = False, draw_figures: bool = False, return_scores: bool = False,
                measure_only: bool = False
                ) -> None | Tuple[str, pd.Series, int, int, pd.Series, pd.Series, pd.Series, str, bool] | Tuple:
        """
        Explain for filter operation

        :param schema: dictionary with new columns names, in case {'col_name': 'i'} will be ignored in the explanation
        :param attributes: only this attributes will be included in the explanation calculation
        :param top_k: top k explanations number, default one explanation only.
        :param show_scores: show scores on explanation
        :param figs_in_row: number of explanations figs in one row
        :param title: explanation title
        :param use_sampling: whether to use sampling or not
        :param sample_size: the size of the sample to use. Can be a percentage of the dataframe size if below 1. Default is 5000.
        :param debug_mode: Developer option. Disables multiprocessing for easier debugging. Default is False. Can possibly add more debug options in the future.
        :param draw_figures: Whether or not to draw the figures in this stage. Defaults to True. Set to False if you want to draw the figures later.
        :param return_scores: Whether or not to return the scores. Defaults to False.

        :return: explain figures
        """

        if explainer == 'shapley':
            measure = ShapleyMeasure()
            # else: score = 0
            top_fact, facts = measure.get_most_contributing_fact(self, self.left_df, self.right_df, self.attribute,
                                                                 None, consider=consider, top_k=top_k, cont=cont,
                                                                 att=attr)
            # exp = f'The result of joining dataframes \'{self.left_name}\' and \'{self.right_name}\' on attribute \'{self.attribute}\' is not empty.\nThe following fact from dataframe \'{self.left_name}\' has significantly contributed to this result:\n'
            # facts = list({k: v for k, v in sorted(facts.items(), key=lambda item: -item[1])}.items())
            # fact_idx = 0
            # exp += str(facts[fact_idx][0])
            # top_k -= 1
            # while top_k > 0:
            #     fact_idx += 1
            #     exp += '\n and then\n'
            #     exp += str(facts[fact_idx][0])
            #     top_k -= 1
            return None
        if attributes is None:
            attributes = []

        if schema is None:
            schema = {}

        backup_left_df, backup_right_df, backup_res_df, combined_source_df = None, None, None, None
        if use_sampling:
            backup_left_df, backup_right_df, backup_res_df = self.left_df, self.right_df, self.result_df
            self.left_df, self.right_df, self.result_df = self.sample(self.left_df, sample_size), self.sample(
                self.right_df, sample_size), self.sample(self.result_df, sample_size)
            # If sampling is used, calc_measure needs an unsampled source_df to create bins for its score dict, otherwise
            # the explanation creation will be affected by the sampling, and not just the measure calculation.
            # Therefore, we concat the left and right DataFrames to create the unsampled source DataFrame.
            combined_source_df = pd.concat([self.left_df, self.right_df], axis=0)

        # When using the FEDEx explainer, the exceptionality measure is used to calculate the explanation.
        measure = ExceptionalityMeasure()
        self._measure = measure
        scores = measure.calc_measure(self, schema, attributes, unsampled_source_df=combined_source_df,
                                      unsampled_res_df=backup_res_df, debug_mode=debug_mode)

        if use_sampling:
            self.left_df, self.right_df, self.result_df = backup_left_df, backup_right_df, backup_res_df

        if measure_only:
            return scores

        ret_val = measure.calc_influence(utils.max_key(scores), top_k=top_k, figs_in_row=figs_in_row,
                                      show_scores=show_scores, title=title, debug_mode=debug_mode,
                                      draw_figures=draw_figures)

        if return_scores:
            return ret_val, scores
        else:
            return ret_val, None

    def draw_figures(self, title: str, scores: pd.Series, K: int, figs_in_row: int, explanations: pd.Series, bins: pd.Series,
                      influence_vals: pd.Series, source_name: str, show_scores: bool,
                     added_text: dict | None = None, added_text_name: str | None = None) -> tuple:
        return self._measure.draw_figures(
            title=title, scores=scores, K=K, figs_in_row=figs_in_row,
            explanations=explanations, bins=bins, influence_vals=influence_vals,
            source_name=source_name, show_scores=show_scores, added_text=added_text,
            added_text_name=added_text_name
        )



import numpy as np
import pandas as pd
from numpy import number
from pandas import Series
from scipy import stats
from typing import Callable

from fedex_generator.Measures.BaseMeasure import BaseMeasure, START_BOLD, END_BOLD
from fedex_generator.Measures.Bins import Bin, MultiIndexBin
from fedex_generator.commons import utils

# A constant dictionary that maps operation names to their corresponding functions, for convenience and clarity
OP_TO_FUNC = {
    'count': np.sum,
    'sum': np.sum,
    'max': np.max,
    'min': np.min,
    'mean': np.mean,
    'prod': np.prod,
    'sem': stats.sem,
    'var': np.var,
    'std': np.std,
    'median': np.median,
    'first': lambda bin_values: bin_values[0],
    'last': lambda bin_values: bin_values[-1],
    'size': np.sum,
    'nunique': lambda bin_values: len(set(bin_values))
}


def draw_bar(x: list, y: list, avg_line=None, items_to_bold=None, head_values=None, xname=None, yname=None, alpha=1.,
             ax=None):
    """
    Draw a bar chart with optional features.

    :param x: List of x-axis values.
    :param y: List of y-axis values.
    :param avg_line: Optional; a value to draw a horizontal line representing the average.
    :param items_to_bold: Optional; list of items to highlight in the bar chart.
    :param head_values: Optional; list of values to display above each bar.
    :param xname: Optional; label for the x-axis.
    :param yname: Optional; label for the y-axis.
    :param alpha: Optional; transparency level of the bars (default is 1.0).
    :param ax: Optional; matplotlib axes object to draw the bar chart on.
    """

    width = 0.5
    ind = np.arange(len(x))

    # Convert the x and y values to valid LaTeX strings, if needed
    x = x if utils.is_numeric(x) else [utils.to_valid_latex(i) for i in x]
    y = y if utils.is_numeric(y) else [utils.to_valid_latex(i) for i in y]

    # Convert the items to bold to valid LaTeX strings, if needed
    if items_to_bold is not None:
        items_to_bold = items_to_bold if utils.is_numeric(items_to_bold) else [utils.to_valid_latex(i) for i in
                                                                               items_to_bold]

    # Create the bar chart
    bar = ax.bar(ind, y, width, alpha=alpha)
    ax.set_xticks(ind)
    ax.set_xticklabels(tuple([str(i) for i in x]), rotation='vertical')
    ax.set_ylim(min(y) - min(y) * 0.01, max(y) + max(y) * 0.001)

    # Add the average line, if provided
    if avg_line is not None:
        ax.axhline(avg_line, color='red', linewidth=1)

    # Highlight the items to bold, if provided
    if items_to_bold is not None:
        for item in items_to_bold:
            bar[x.index(item)].set_color('green')

    # Add the head values, if provided
    if head_values is not None:
        for i, col in enumerate(bar):
            yval = col.get_height()
            ax.text(col.get_x(), yval + .05, head_values[i])

    # Add the x and y axis labels, if provided
    if xname is not None:
        ax.set_xlabel(utils.to_valid_latex_with_escaped_dollar_char(xname), fontsize=24)

    if yname is not None:
        ax.set_ylabel(utils.to_valid_latex_with_escaped_dollar_char(yname), fontsize=24)


def flatten_other_indexes(series, main_index):
    """
    Flatten all indexes of a pandas Series except the specified main index.

    This function resets the specified main index of the Series, flattens the remaining indexes into a single index,
    and then sets the main index back as an additional level of the index.

    :param series: The input pandas Series with a MultiIndex.
    :param main_index: The name of the main index to retain.
    :return: A pandas Series with the main index retained and other indexes flattened.
    """
    df = pd.DataFrame(series)
    df = df.reset_index(main_index)
    index_name = "_".join([ind for ind in df.index.names])
    df.index = df.index.to_flat_index()
    df.index.names = [index_name]
    df = df.set_index(main_index, append=True)
    return df[df.columns[0]]


class DiversityMeasure(BaseMeasure):
    """
    A class implementing the diversity measure, a measure for interestingness of GroupBy operations.\n
    This measure is described in the article "FEDEX: An Explainability Framework for Data Exploration Steps" by
    Daniel Deutch, Amir Gilad, Tova Milo, Amit Mualem, Amit Somech.\n
    From the article: "intuitively, a group-by step that yields a dataframe with a highly diverse set of aggregated values,
    implies a large difference between the groups".\n
    The measure is defined as the coefficient of variation: \n
    .. math:: I_A (d_{in}, q_g, d_{out}) = CV(d_{out}[A]) = \\frac{1}{\\tilde{a}} * \\sqrt{\\frac{\\sum_i (a_i - \\tilde{a})^2}{n}}\n
    Where:\n
    - :math:`I_A` is the diversity measure for attribute A.\n
    - :math:`d_{in}` is the input dataframe.\n
    - :math:`q_g` is the query, usually a group-by operation.\n
    - :math:`d_{out}` is the output dataframe.\n
    - :math:`A` is the attribute of interest.\n
    - :math:`CV` is the coefficient of variation.\n
    - :math:`\\tilde{a}` is the mean of the values of attribute A.\n
    - :math:`a_i` is the i-th value of attribute A in the output dataframe.\n
     """

    MAX_BARS = 25

    def __init__(self):
        super().__init__()

    def get_agg_func_from_name(self, name: str) -> str | Callable:
        """
        Retrieve the aggregation function corresponding to the given name.

        This function attempts to find the aggregation function based on the provided name.
        It first checks if the name corresponds to a method of `pd.Series`. If not, it checks
        if the name is in the `OP_TO_FUNC` dictionary. If the function is still not found,
        it searches within the `agg_dict` of the `operation_object`.

        :param name: The name of the aggregation function to retrieve.
        :return: The aggregation function corresponding to the given name.
        """

        # Get the operation name from the column name.
        # Name can be an int or a float in some cases, so we need to convert it to a string.
        if isinstance(name, int) or isinstance(name, float):
            name = str(name)
        # It can also be a string, or in the case of a multi-index, a tuple.
        if isinstance(name, str):
            operation = name.split("_")[-1].lower()
        elif isinstance(name, tuple):
            name = "_".join(x for x in name)
            operation = name.lower()
        else:
            raise TypeError(
                f"The type of the column name is {type(name)}, which we the developers did not expect and do not know how to handle. If you are a developer, please fix this. If you are a user, please report this. Thank you.")

        # Check if the name corresponds to a method of `pd.Series` or is in the `OP_TO_FUNC` dictionary
        # If the operation is a method of `pd.Series`, we can quickly return the function. Unless the operation name coincides with a column name,
        # in which case we can't be sure if it's a method or a column name, so we need to check the agg_dict.
        # As an example: the spotify dataset has a column named "mode", which is in-fact the reason why this condition
        # was discovered as necessary, because reaching here with "mode" as the operation name would cause an aggregation failure
        # exception in the calling function once it tries to use the function returned here.
        if hasattr(pd.Series, operation) and not operation in self.operation_object.result_df.columns:
            # In cases like 'size', this can end up returning a property, which will cause an exception when it's used
            # in the calling function. So we need to check if it's callable. Otherwise we'll find it later in the
            # agg_dict.
            attr = getattr(pd.Series, operation)
            if callable(attr):
                return attr
        elif operation in OP_TO_FUNC:
            return OP_TO_FUNC[operation]

        res = []
        # Search within the `agg_dict` of the `operation_object`.
        for x in self.operation_object.agg_dict:
            op = self.operation_object.agg_dict[x]
            if not isinstance(op, str):
                for y in op:
                    res.append(x + '_' + (y if isinstance(y, str) else y.__name__))
            else:
                res.append(x + '_' + op)
        # Note that this can fail if the agg_dict it not set properly, or if the name is not in the agg_dict.
        # For example, if the column is "All", meaning all columns are aggregated, then the name will not be in the
        # agg_dict, and this will fail.
        try:
            aggregation_index = res.index(name)
            return list(self.operation_object.agg_dict.values())[0][aggregation_index]
        except (ValueError, IndexError) as e:
            if any([x.startswith("All_") for x in res]):
                return list(self.operation_object.agg_dict.values())[0][0]
            # It is also possible that we get names that don't include the aggregation function, but are still
            # valid, like "col_name", and res will have "col_name_mean" in it.
            # This won't be caught anywhere above, so we check for it here.
            contains_name = [x for x in res if x.startswith(name)]
            if len(contains_name) > 0:
                agg_funcs = [x.split("_")[-1] for x in contains_name]
                # We return a random function from the list of aggregation functions, since we don't know which one to use.
                return np.random.choice(agg_funcs)

            raise e
        # aggregation_index = res.index(name)
        # return list(self.operation_object.agg_dict.values())[0][aggregation_index]

    def draw_bar(self, bin_item: MultiIndexBin, influence_vals: dict = None, title=None, ax=None, score=None,
                 show_scores: bool = False, explanation_num: int | None = None):
        """
        Draw a bar chart for a given bin item with optional features.

        This function creates a bar chart for the provided `bin_item`, highlighting the most influential values
        and optionally displaying additional information such as the average line, title, and scores.

        :param bin_item: The `MultiIndexBin` object containing the bin data to be plotted.
        :param influence_vals: Optional; a dictionary of influence values for the bin items.
        :param title: Optional; the title of the bar chart.
        :param ax: Optional; the matplotlib axes object to draw the bar chart on.
        :param score: Optional; the score to be displayed in the title if `show_scores` is True.
        :param show_scores: Optional; a boolean indicating whether to display the score in the title (default is False).
        :param explanation_num: Optional; an integer representing the explanation number to be included in the title.
        """
        try:
            # Get the index of the maximum value and its influence
            max_values, max_influence = self.get_max_k(influence_vals, 1)
            max_value = max_values[0]

            # Calculate the average value of the result column
            res_col = bin_item.get_binned_result_column()
            average = float(utils.smart_round(res_col.mean()))

            # Get the aggregation function corresponding to the result column name
            agger_function = self.get_agg_func_from_name(res_col.name)
            aggregated_result = res_col.groupby(bin_item.get_bin_name()).agg(agger_function)

            # Flatten the other indexes of the result column
            bin_result = bin_item.result_column.copy()
            bin_result = flatten_other_indexes(bin_result, bin_item.get_bin_name())
            smallest_multi_bin = MultiIndexBin(bin_item.source_column, bin_result, 0)
            # Calculate the influence values for the bin items
            influences = self.get_influence_col(res_col, smallest_multi_bin, True)
            rc = res_col.reset_index([bin_item.get_bin_name()])

            # Get the relevant items and influences for the maximum value
            relevant_items = rc[rc[bin_item.get_bin_name()] == max_value]
            relevant_influences = dict([(k, influences[k]) for k in relevant_items.index])
            # Get the top 10 most influential values
            max_values, max_influence = self.get_max_k(relevant_influences, 10)
            max_values = sorted(max_values)
            labels = set(aggregated_result.keys())

            # Limit the number of bars to 25, and sort the labels
            if len(labels) > self.MAX_BARS:
                top_items, _ = self.get_max_k(influence_vals, self.MAX_BARS)
                labels = sorted(top_items)
            else:
                labels = sorted(labels)

            # Get the aggregated values for the labels, and draw the bar chart
            aggregate_column = [aggregated_result.get(item, 0) for item in labels]
            if explanation_num is not None:
                title = f"{START_BOLD}[{explanation_num}]{END_BOLD} {title}" if title else f"{START_BOLD}[{explanation_num}]{END_BOLD}"
            if show_scores:
                ax.set_title(f'score: {score}\n{utils.to_valid_latex(title)}', fontdict={'fontsize': 10})
            else:
                ax.set_title(utils.to_valid_latex(title), fontdict={'fontsize': 20})

            draw_bar(labels, aggregate_column, aggregated_result.mean(), [max_value],
                     xname=f'{bin_item.get_bin_name()} values', yname=bin_item.get_value_name(),
                     ax=ax)
            ax.set_axis_on()

            # Rotate the x-axis labels for better readability, if there are over 8 labels
            if len(labels) > 8:
                ax.tick_params(axis='x', rotation=45, tickdir='in')

        except Exception as e:
            # In the case of an exception, draw a bar chart using the draw_bar method defined outside the class
            columns = bin_item.get_binned_result_column()
            max_group_value = self._find_max_group_value(columns, max_value)

            # If there are more than 25 bars, limit the number of bars to 25
            if len(columns) > self.MAX_BARS:
                columns = self._select_top_columns(influence_vals, columns, max_value, max_group_value)

            title = self._fix_explanation(title, columns, max_value, max_group_value)
            if explanation_num is not None:
                title = f"{START_BOLD}[{explanation_num}]{END_BOLD} {title}" if title else f"{START_BOLD}[{explanation_num}]{END_BOLD}"
            ax.set_title(utils.to_valid_latex(title), fontdict={'fontsize': 20})

            draw_bar(list(columns.index),
                     list(columns),
                     average,
                     [max_group_value] if not isinstance(max_group_value, list) else max_group_value,
                     yname=bin_item.get_bin_name(),
                     ax=ax,
                     )

            # Rotate the x-axis labels for better readability, if there are over 8 labels
            if len(columns) > 8:
                ax.tick_params(axis='x', rotation=45, tickdir='in')
            ax.set_axis_on()


    def _find_max_group_value(self, columns, max_value) -> str | list:
        """
        Finds the group value corresponding to the max value in the columns.
        :param columns: The columns to search for the max value.
        :param max_value: The max value to search for in the columns.
        :return: The group value corresponding to the max value.
        """
        # Using a simple columns == max_value works, so long as the index is not a multi-index one.
        # Otherwise, the max_value may end up being from one of the index levels, which will cause an exception as
        # it is not in the columns.
        if isinstance(columns.index, pd.MultiIndex) and not isinstance(max_value, tuple):
            # Search the index for any match with the max value, even if it's from a subset of the index
            # For example, if we have a MultiIndex with levels A,B,C, and the max value is from B, we can still
            # find it in the index.
            index = columns.index.to_frame()
            # Get all locations in the index where the max_value is found
            indexes = set()
            for col in index.columns:
                col_indexes = index[index[col] == max_value].index
                if len(col_indexes) > 0:
                    indexes.update(col_indexes)
            # Get the columns corresponding to the max value
            indexes = list(indexes)
            max_group_value = list(columns.loc[indexes].to_dict().keys())
            # We set this to also be equal, since this value will get accessed again in _fix_explanation, and it
            # should match the way we fixed it here.
            max_value = max_group_value
        else:
            max_group_value = list(columns[columns == max_value].to_dict().keys())[0]

        return max_group_value


    def _select_top_columns(self, influence_vals, columns, max_value, max_group_value) -> pd.Series:
        """
        Select the top columns to display in the bar chart.
        :param influence_vals: The influence values of the columns.
        :param columns: The columns to select the top items from.
        :param max_value: The max value found.
        :param max_group_value: The name of the group corresponding to the max value.
        :return: The top columns to display in the bar chart.
        """
        top_items, _ = self.get_max_k(influence_vals, self.MAX_BARS)
        # Just like before, we need different handling for multi-index cases, in the case where there are
        # multiple top items, but they are not tuples, i.e. they are not keys for the multi-index
        if isinstance(columns.index, pd.MultiIndex) and not all([isinstance(i, tuple) for i in top_items]):
            index = columns.index.to_frame()
            # We use a list this time, because we need to maintain the same order as the top items
            indexes = []
            for item in top_items:
                for col in index.columns:
                    col_indexes = index[index[col] == item].index
                    if len(col_indexes) > 0:
                        indexes.extend(col_indexes)
            if len(indexes) > len(columns):
                indexes = indexes[:len(columns)]
        else:
            indexes = []
            # Get the indexes of the top items, in order of the top items
            for item in top_items:
                indexes.extend(columns[columns == item].index)
        columns_backup = columns.copy()
        # Using the indexes in this way will automatically sort the columns such that their order matches that
        # of the top items
        columns = columns.loc[indexes]
        columns = columns[:self.MAX_BARS]
        cols_dtype_is_numeric = utils.is_numeric(columns)

        # If the max value is not in the columns, add it to the columns
        if not isinstance(max_group_value, list):
            if max_group_value not in columns:
                columns = columns.append(pd.Series(max_value, index=[max_group_value]))
        else:
            # Not having the index sorted can raise a performance warning, so we sort it if it's not sorted
            if not columns.index._is_lexsorted():
                columns = columns.sort_index()
            for value in max_group_value:
                if value not in columns:
                    # If the max value is the same dtype as the columns, we can append it directly
                    if cols_dtype_is_numeric and type(value) in (int, float):
                        columns = columns._append(pd.Series(max_value, index=[value]))
                    # Else, retrieve the value from the columns_backup, which is guaranteed to have the value
                    else:
                        value_to_add = columns_backup.loc[value]
                        columns = columns._append(pd.Series(value_to_add, index=[value]))

        return columns

    @staticmethod
    def _fix_explanation(explanation: str, binned_column: Series, max_value, max_group_value) -> str:
        """
        Change explanation column to match group by
        :param explanation:  bin explanation
        :param binned_column: bin column
        :param max_value: max value

        :return: new explanation
        """
        # Tuple is used to represent multi-index columns
        # if isinstance(max_value, tuple):
        #     max_group_value = list(binned_column.loc[max_value].to_dict().keys())[0]
        # else:
        #     max_group_value = list(binned_column[binned_column == max_value].to_dict().keys())[0]
        binned_column_name = str(binned_column.name)
        max_value_name = binned_column_name.replace('_', '\\ ')
        try:
            max_group_value.replace('$', '\\$')
            max_value_name.replace('$', '\\$')
        except:
            pass
        group_by_name = binned_column.to_frame().axes[0].name

        return explanation.replace(f'\'{max_value_name}\'=\'{max_value}\'',
                                   f'\'{group_by_name}\' = \'{max_group_value}\'')


    def _try_fix_explanation(self, explanation: str, binned_column: MultiIndexBin, max_value) -> str:
        columns = binned_column.get_binned_result_column()
        max_group_value = self._find_max_group_value(columns, max_value)
        return self._fix_explanation(explanation, columns, max_value, max_group_value)



    def interestingness_only_explanation(self, source_col: Series, result_col: Series, col_name: str) -> str:
        return f"After employing the GroupBy operation we can see highly diverse set of values in the column '{col_name}'\n" \
               f"The variance" + \
            (f" was {self.calc_var(source_col)} and now it " if source_col is not None else "") + \
            f" is {self.calc_var(result_col)}"

    def calc_influence_col(self, current_bin: Bin):
        # Get the values of the current bin
        bin_values = current_bin.get_bin_values()
        source_col = current_bin.get_source_by_values(bin_values)
        res_col = current_bin.get_result_by_values(bin_values)
        # Calculate the diversity score for the current bin
        score_all = self.calc_diversity(source_col, res_col)

        influence = []
        # Compute the diversity score for each value in the bin, then calculate the influence of each value
        # as the difference between the score of all values and the score without the current value
        for value in bin_values:
            source_col_only_list = current_bin.get_source_by_values([b for b in bin_values if b != value])
            res_col_only_list = current_bin.get_result_by_values([b for b in bin_values if b != value])

            score_without_bin = self.calc_diversity(source_col_only_list, res_col_only_list)
            influence.append(score_all - score_without_bin)

        return influence

    def calc_var(self, pd_array) -> number:
        """
        Calculate the variance of a pandas array.

        This method computes the variance of the given pandas array. If the array contains numeric values,
        it returns the variance of those values. If the array contains non-numeric values, it calculates
        the variance based on the frequency of each unique value in the array.

        :param pd_array: The pandas array for which to calculate the variance.
        :return: The variance of the array.
        """
        if utils.is_numeric(pd_array):
            return np.var(pd_array)

        appearances = (pd_array.value_counts()).to_numpy()
        mean = np.mean(appearances)
        return np.sum(np.power(appearances - mean, 2)) / len(appearances)

    def calc_diversity(self, source_col: Series, res_col: Series) -> float:
        """
        Calculate the diversity measure for the given source and result columns.

        This method computes the diversity measure, defined as the coefficient of variation,
        for the provided source and result columns. The diversity measure is calculated as
        the ratio of the variance of the result column to the variance of the source column - see the class description
        for the formula.

        :param source_col: The source column as a pandas Series. This can be None.
        :param res_col: The result column as a pandas Series.
        :return: The diversity measure as a float.
        """
        var_res = self.calc_var(res_col)

        if source_col is not None:
            var_rel = self.calc_var(source_col)
        else:
            var_rel = 1.

        res = (var_res / var_rel) if var_rel != 0 else (0. if var_res == 0 else var_res)
        return 0 if np.isnan(res) else res

    def calc_measure_internal(self, bin: Bin):
        result_column = bin.get_binned_result_column().dropna()
        if bin.name == "MultiIndexBin":
            operation = self.get_agg_func_from_name(result_column.name)
            result_column = result_column.groupby(bin.get_bin_name()).agg(operation)
        return self.calc_diversity(None if bin.source_column is None else bin.source_column.dropna(),
                                   result_column)

    def build_operation_expression(self, source_name):
        from fedex_generator.Operations.GroupBy import GroupBy
        if isinstance(self.operation_object, GroupBy):
            return f'{source_name}.groupby({self.operation_object.group_attributes})' \
                   f'.agg({self.operation_object.agg_dict})'

    def build_explanation(self, current_bin: Bin, max_col_name, max_value, source_name):
        res_col = current_bin.get_binned_result_column()
        # If the result column is categorical, return an empty explanation
        if utils.is_categorical(res_col):
            return ""

        var = self.calc_var(res_col)
        # If the current bin is numeric or categorical, set the max value to the provided value
        if current_bin.name == "NumericBin" or current_bin.name == 'CategoricalBin':
            max_value_numeric = max_value

        # Otherwise, some additional processing is required
        elif current_bin.name == "MultiIndexBin":
            # Get the result column values for the maximum value, then calculate the aggregation function
            # on those values and set the max value to the result of the aggregation function
            bin_values = current_bin.get_result_by_values([max_value])
            operation = self.get_agg_func_from_name(res_col.name)
            # If the operation is a callable function, call it with the bin values
            if callable(operation):
                max_value_numeric = operation(bin_values)
            # Otherwise, it is a string, so we need to get the function from the OP_TO_FUNC dictionary
            else:
                operation = OP_TO_FUNC[operation]
                max_value_numeric = operation(bin_values)

            # Get the name of the column with the maximum value, and compute the variance of the result column
            max_col_name = current_bin.get_bin_name()
            res_col = res_col.groupby(max_col_name).agg(operation)
            var = self.calc_var(res_col)

        elif current_bin.name == "NoBin":
            result_column = current_bin.get_binned_result_column()
            max_value_numeric = max_value
            max_value = result_column.index[result_column == max_value].tolist()
            max_col_name = result_column.index.name

        elif type(current_bin) == Bin:
            raise Exception("Bin is not supported")
        else:
            raise Exception(f"unknown bin type {current_bin.name}")

        # Get the standard deviation of the result column and calculate the z-score of the maximum value,
        # determining whether the value is relatively high or low, and generate the explanation
        sqr = np.sqrt(var)
        x = ((max_value_numeric - np.mean(res_col)) / sqr) if sqr != 0 else 0
        group_by_text = utils.to_valid_latex_with_escaped_dollar_char(f"'{max_col_name}'='{max_value}'", True)
        proportion = 'low' if x < 0 else 'high'
        proportion_column = utils.to_valid_latex_with_escaped_dollar_char(f"{proportion} '{res_col.name}'", True)

        expl = f"Groups with {START_BOLD}{group_by_text}{END_BOLD}(in green)\n" \
               f"have a relatively {START_BOLD}{proportion_column}{END_BOLD} value:\n" \
               f"{utils.smart_round(np.abs(x))} standard deviations {proportion}er than the mean\n" \
               f"({utils.smart_round(np.mean(res_col))})"

        return expl



import inspect
import random
import numpy as np
import pandas as pd
import scipy
import statistics
import operator
from enum import Enum
from fedex_generator.commons.kstest import *


def to_valid_latex(string, is_bold: bool = False) -> str:
    """
    Convert a string to a valid latex string.
    :param string: The input string.
    :param is_bold: Whether the string should be bold.
    :return: The latex string.
    """
    latex_string = str(string)  # unicode_to_latex(string)
    # Choose the space character based on whether the string should be bold.
    space = r'\ ' if is_bold else ' '
    # Escape special characters and replace spaces with the chosen space character.
    final_str = latex_string.replace("&", "\\&").replace("#", "\\#").replace(' ', space).replace("_", space)
    return final_str


def to_valid_latex_with_escaped_dollar_char(string, is_bold: bool = False) -> str:
    """
    Convert a string to a valid latex string.
    This function adds the $ character to the list of characters that are escaped, while the function to_valid_latex
    does not escape the $ character. Other than that, the two functions are identical.
    :param string: The input string.
    :param is_bold: Whether the string should be bold.
    :return: The latex string.
    """
    return to_valid_latex(string, is_bold).replace("$", "\\$")


def smart_round(number) -> str:
    """
    Round a number to a set, certain number of decimal places.
    :param number: The input number.
    :return: The rounded number, as a string.
    """
    return f"{number:.1f}" if 0 <= number <= 2 else f"{number:.0f}"


def format_bin_item(item) -> str:
    """
    Formats a bin item to a string.
    :param item: The input item.
    :return: The formatted item.
    """
    # If the item is categorical or a date, return it as a string directly.
    if is_categorical(item) or is_date(item):
        return str(item)

    # If the item is an integer, return it as a string after converting it to int.
    if hasattr(item, "is_integer") and item.is_integer():
        return str(int(item))

    # If the item is a float between 1 and -1, return it as a string rounded to 2 decimal places.
    if -1 < item < 1:
        return f"{item:.2f}"

    # Else, return the item as a string rounded to 1 decimal place.
    return f"{item:.1f}"


def max_key(d) -> str:
    """
    Get the key with the maximum value from a dictionary.
    :param d: The input dictionary.
    :return: The key with the maximum value.
    """
    return max(d.items(), key=operator.itemgetter(1))[0]


def is_name_informative(name) -> bool:
    """
    Check if a name is informative, i.e. it includes at least one letter.
    :param name: The input name.
    :return: Whether the name is informative.
    """
    import string
    return any([char in string.ascii_letters for char in name])


def get_calling_params_name(item) -> str:
    """
    Get the name of the variable that called the function.
    :param item: The input item.
    :return: The name of the variable that called the function.
    """
    # Get the stack frames.
    frames = inspect.stack()
    highest_var = None
    # Iterate over the frames, starting from the third frame, looking for the variable that called the function
    # by comparing the id (memory address) of the variable with the id of the variables in the previous frames.
    for frame in frames[2:]:
        prev_locals = frame[0].f_locals
        for var_name in prev_locals:
            if id(item) == id(prev_locals[var_name]) and is_name_informative(var_name):
                highest_var = var_name

    return highest_var


# Define known numeric and categorical types.
NUMERIC_TYPES = ['float64', 'int32', 'int64', 'int', 'datetime64[ns]']
CATEGORICAL_TYPES = ['category', 'object', 'bool', 'tuple']


def drop_nan(array) -> np.ndarray:
    """
    Drop NaN values from an array.
    :param array: The input array.
    :return: The array with NaN values dropped, as a flattened numpy array.
    """
    return pd.Series(array).dropna().to_numpy().flatten()


class ArrayType(Enum):
    """
    Enum class to represent the type of an array.
    Either Categorical, Numeric, or Unknown.
    """
    Categorical = 1
    Numeric = 2
    Unknown = 3


def get_array_type(array_like) -> ArrayType:
    """
    Get the type of an array, by the ArrayType enum: Categorical, Numeric, or Unknown.
    :param array_like: The input array.
    :return: The type of the array.
    """
    # If the array is of a known categorical type, return Categorical.
    if np.array(array_like).dtype.name in CATEGORICAL_TYPES or np.array(array_like).dtype.name.startswith('str'):
        return ArrayType.Categorical

    # If the array is of a known numeric type, return Numeric.
    if np.array(array_like).dtype.name in NUMERIC_TYPES:
        return ArrayType.Numeric

    # Else, return Unknown.
    return ArrayType.Unknown


def is_categorical(array_like) -> bool:
    """
    Check if an array is categorical.
    :param array_like: The input array.
    :return: Whether the array is categorical.
    """
    array_type = get_array_type(array_like)
    if array_type == ArrayType.Categorical:
        return True

    if array_type == ArrayType.Numeric:
        return False

    # Throw an error if the array type is unknown.
    raise RuntimeError(f"Bad type: {np.array(array_like).dtype.name}")


def is_numeric(array_like) -> bool:
    """
    Check if an array is numeric.
    :param array_like: The input array.
    :return: Whether the array is numeric.
    """
    return not is_categorical(array_like)


def is_date(item) -> bool:
    """
    Check if an item is a date.
    :param item: The input item.
    :return: Whether the item is a date
    """
    if "date" in str(type(item)):
        return True

    return hasattr(item, 'date') or hasattr(item, 'time')



from typing import List
import pandas as pd
import numpy as np

from fedex_generator.commons.consts import TOP_K_DEFAULT, DEFAULT_FIGS_IN_ROW

SAMPLE_SIZE = 5000
RANDOM_SEED = 42


class Operation:
    """
    An abstract class for operations within the FEDEx explainability framework.\n
    All implemented operations should inherit from this class.
    """

    def __init__(self, scheme: dict):
        """
        :param scheme: the scheme of the dataset
        """
        self.scheme = scheme
        self.bins_count = 500

    def set_bins_count(self, n: int) -> None:
        """
        Set the number of bins used when explaining the operation.
        :param n: the number of bins
        """
        self.bins_count = n

    def get_bins_count(self) -> int:
        """
        :return: The number of bins used when explaining the operation.
        """
        return self.bins_count

    def explain(self, schema: dict = None, attributes: List[str] = None, top_k: int = TOP_K_DEFAULT,
                figs_in_row: int = DEFAULT_FIGS_IN_ROW, show_scores: bool = False, title: str = None,
                corr_TH: float = 0.7, use_sampling: bool = True):
        """
        Explain for operation
        :param schema: dictionary with new columns names, in case {'col_name': 'i'} will be ignored in the explanation
        :param attributes: only this attributes will be included in the explanation calculation
        :param top_k: top k explanations number, default one explanation only.
        :param figs_in_row: number of explanations figs in one row
        :param show_scores: show scores on explanation
        :param title: explanation title


        :return: explain figures
        """
        raise NotImplementedError()

    def draw_figures(self, title: str, scores: pd.Series, K: int, figs_in_row: int, explanations: pd.Series, bins: pd.Series,
                      influence_vals: pd.Series, source_name: str, show_scores: bool,
                     added_text: dict | None = None, added_text_name: str | None = None) -> None:
        """
        Draws the figures for the explanations.
        :param title: The title of the plot.
        :param scores: The scores of the attributes.
        :param K: The number of top attributes to consider.
        :param figs_in_row: The number of figures to display in a row.
        :param explanations: The explanations generated for the top K attributes.
        :param bins: The bins of the attributes.
        :param influence_vals: The influence values of the attributes.
        :param source_name: The name of the source DataFrame.
        :param show_scores: Whether to show the scores on the plot.
        :param added_text: Additional text to add to the bottom of each figure. Optional. A dict with explanation as key, and a sub-dict with 'text' and 'position' as keys.
        :param added_text_name: The name of the additional text to be displayed. Optional. If provided, it will be displayed above the added text.
        """
        raise NotImplementedError()

    def present_deleted_correlated(self, figs_in_row: int = DEFAULT_FIGS_IN_ROW):
        """
        Present the attributes that were deleted due to high correlation.
        :param figs_in_row: number of explanations figs in one row.
        """
        return NotImplementedError()

    def sample(self, df: pd.DataFrame, sample_size: int | float = SAMPLE_SIZE) -> pd.DataFrame:
        """
        Uniformly sample the dataframe to the given sample size.
        :param df: The dataframe to sample.
        :param sample_size: The sample size to use. Default is SAMPLE_SIZE. If the sample size is below 1,
         it is considered a percentage of the dataframe size.
        :return: The sampled dataframe.
        """
        # If the sample size is below 1, we consider it to be a percentage of the dataframe size.
        if sample_size <= 0:
            raise ValueError("Sample size must be a positive number.")
        if 0 < sample_size < 1:
            sample_size = int(df.shape[0] * sample_size)
        # If the sample size is below the default sample size, we use the default sample size.
        # We do this because we know, from empirical testing, that our default size is a good balance between
        # performance and accuracy, and going below that size can lead to too big a loss in accuracy for
        # a very small gain in performance.
        if sample_size < SAMPLE_SIZE:
            sample_size = SAMPLE_SIZE
        # If the sample size is larger than the dataframe, we return the dataframe as is
        if df.shape[0] <= sample_size:
            return df
        else:
            # Convert the sample size to an integer, in case it was passed as a float above 1.
            sample_size = int(sample_size)
            # We use a set seed so that the user will always get the same explanation when using sampling.
            generator = np.random.default_rng(RANDOM_SEED)
            uniform_indexes = generator.choice(df.index, sample_size, replace=False)
            return df.loc[uniform_indexes]



import os

from setuptools import setup, find_packages


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version():
    for line in read('src/fedex_generator/__init__.py').splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def get_long_description():
    with open('README.md', 'r') as fh:
        return fh.read()


setup(
    name='fedex_generator',
    version='1.0.6.0',#get_version(),
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    long_description_content_type="text/markdown",
    long_description=get_long_description(),  # Long description read from the readme file
    project_urls={
        'Git': 'https://github.com/analysis-bots/FEDEx_Generator',
    },
    install_requires=[
        'wheel',
        'pandas>=2.2.3',
        'numpy>=2.1.3',
        'matplotlib>=3.9.0',
        'scipy>=1.14.1',
        'paretoset>=1.2.4',
    ]
)



from collections.abc import Iterable

import pandas as pd
from typing import List, Any, Hashable

from numpy import number
from pandas import Series
from pandas.core.interchange.dataframe_protocol import DataFrame

from fedex_generator.commons import utils


class Bin(object):
    """
    Base class for all binning methods.
    """

    def __init__(self, source_column: pd.Series, result_column: pd.Series, name: str):
        """
        :param source_column: The source column to bin.
        :param result_column: The result column to bin.
        :param name: The name of the binning method.
        """
        self.source_column, self.result_column = source_column, result_column
        self.name = name

    def get_binned_source_column(self) -> pd.Series | None:
        """
        Get the bin's source column.
        :return: The source column.
        """
        return self.source_column

    def get_binned_result_column(self) -> pd.Series | None:
        """
        Get the bin's result column.
        :return: The result column.
        """
        return self.result_column

    def get_source_by_values(self, values: Iterable | DataFrame | Series | dict) -> pd.Series | None:
        """
        Get values from the source column by the given values.
        :param values: The values to get from the source column.
        :return: A series with the values from the source column, or None if the source column is None.
        """
        source_column = self.get_binned_source_column()
        if source_column is None:
            return None

        return source_column[source_column.isin(values)]

    def get_result_by_values(self, values: Iterable | DataFrame | Series | dict) -> pd.Series:
        """
        Get values from the result column by the given values.
        :param values: The values to get from the result column.
        :return: A series with the values from the result column.
        """
        result_column = self.get_binned_result_column()
        return result_column[result_column.isin(values)]

    def get_bin_values(self) -> List[Any]:
        """
        Get the values of the bin, sorted and without NaN values.
        :return: A list of the values of the bin.
        """
        # Get the source and result columns
        source_col = self.get_binned_source_column()
        source_col = [] if source_col is None else source_col
        res_col = self.get_binned_result_column()

        # Create a list of the unique values in the source and result columns, then sort it
        values = list(set(source_col).union(set(res_col)))
        values.sort()

        # Return the values after dropping any NaN values
        return list(utils.drop_nan(values))

    def get_bin_name(self):
        """
        Get the name of the bin.
        The name of the bin is the name of the result column.
        :return: The name of the bin.
        """
        return self.result_column.name

    def get_bin_representation(self, item: Any) -> str:
        """
        Get a bin representation of an item.
        The representation is the item formatted as a string via the :func:`fedex_generator.commons.utils.format_bin_item` method.

        :param item: The item to format.
        :return: The formatted item
        """
        return utils.format_bin_item(item)


class UserBin(Bin):
    """
    Base class for user-defined binning methods.
    This is an abstract class and should not be instantiated directly.
    """

    def __init__(self, source_column: Series, result_column: Series):
        super().__init__(source_column, result_column, "UserDefined")

    def get_binned_source_column(self):
        raise NotImplementedError()

    def get_binned_result_column(self):
        raise NotImplementedError()

    def get_bin_name(self):
        raise NotImplementedError()

    def get_bin_representation(self, item):
        raise NotImplementedError()


class MultiIndexBin(Bin):
    """
    A binning method for multi-index columns.
    """

    def __init__(self, source_column: Series, result_column: Series, level_index: int):
        """
        :param source_column: The source column to bin.
        :param result_column: The result column to bin.
        :param level_index: The index of the level to bin
        """
        super().__init__(source_column, result_column, "MultiIndexBin")
        self.level_index = level_index

    def get_source_by_values(self, values):
        if self.source_column is None:
            return None

        return self.source_column[self.source_column.index.isin(values, level=self.level_index)]

    def get_result_by_values(self, values):
        return self.result_column[self.result_column.index.isin(values, level=self.level_index)]

    def get_bin_values(self) -> List[Any]:
        """
        :return: A list of the values of the result column at the level index.
        """
        return list(self.result_column.index.levels[self.level_index])

    def get_base_name(self) -> str:
        """
        :return: The name of the index at the level 0.
        """
        return self.result_column.index.names[0]

    def get_bin_name(self) -> str:
        """
        :return: The name of the index at the level index.
        """
        return self.result_column.index.names[self.level_index]

    def get_value_name(self) -> Hashable:
        """
        Gets the name of the value, in this case the name of the result column.
        :return: The name of the result column.
        """
        return self.result_column.name


class NumericBin(Bin):
    """
    A binning method for numeric columns.
    Automatically manages the binning of the columns based on the provided size.
    """

    def __init__(self, source_column: Series, result_column: Series, size: int):
        """
        :param source_column: pd.Series - The column to bin in the source dataframe.
        :param result_column: pd.Series - The column to bin in the result (after an operation) dataframe.
        :param size: int - The number of bins to create.
        """
        # If the source column is not None, bin it using the qcut method, and use the returned bins to bin the result column
        if source_column is not None:
            binned_source_column, bins = pd.qcut(source_column, size, retbins=True, labels=False, duplicates='drop')
            bins = utils.drop_nan(bins)
            if len(bins) > 0:
                bins[0] -= 1  # stupid hack because pandas cut uses (,] boundaries, and the first bin is [,]
            binned_result_column = pd.cut(result_column, bins=bins, labels=False, duplicates='drop')
        # Otherwise, bin the result column only
        else:
            binned_source_column = None
            binned_result_column, bins = pd.qcut(result_column, size, retbins=True, labels=False,
                                                 duplicates='drop')  # .reset_index(drop=True)

        super().__init__(binned_source_column, binned_result_column, "NumericBin")
        self.bins = bins

    def get_binned_source_column(self):
        if self.source_column is None:
            return None

        bins_dict = dict(enumerate(self.bins))
        return self.source_column.map(bins_dict)

    def get_binned_result_column(self):
        bins_dict = dict(enumerate(self.bins))
        return self.result_column.map(bins_dict)

    def get_bin_representation(self, item: number) -> str:
        """
        Get the bin representation of an item.\n
        This is represented as a string with the bin boundaries.

        :param item: The item to format.
        :return: A string representation of the bin.
        """
        item_index = list(self.bins).index(item)
        next_item = self.bins[item_index + 1]
        return f"({utils.format_bin_item(item)}, {utils.format_bin_item(next_item)}]"


class CategoricalBin(Bin):
    """
    A binning method for categorical columns.
    Automatically manages the binning of the columns based on the provided size.
    """

    @staticmethod
    def get_top_k_values(column: Series, k: int) -> List[Any]:
        """
        Static method to get the top k most frequent values in a column.
        :param column: The column to get the top k values from.
        :param k: The number of top values to return.
        :return: A list of the top k values.
        """
        col_values = [(v, k) for (k, v) in column.value_counts().items()]
        col_values.sort(reverse=True)

        return [v for (k, v) in col_values[:k]]

    def __init__(self, source_column: Series, result_column: Series, size: int):
        """
        :param source_column: The source column to bin.
        :param result_column: The result (after an operation on the source) column to bin.
        :param size: The number of bins to create.
        """
        self.result_values = self.get_top_k_values(result_column, 11 - 1)
        id_map_top_k = dict(zip(self.result_values, self.result_values))

        if source_column is not None:
            source_column.map(id_map_top_k)
        result_column.map(id_map_top_k)
        super().__init__(source_column, result_column, "CategoricalBin")

    def get_source_by_values(self, values):
        source_column = self.get_binned_source_column()
        if source_column is None:
            return None

        return source_column[source_column.isin(values) | source_column.isnull()]

    def get_result_by_values(self, values):
        result_column = self.get_binned_result_column()
        return result_column[result_column.isin(values) | result_column.isnull()]

    def get_bin_values(self):
        return self.result_values


class NoBin(Bin):
    """
    A bin for when no binning is required, but the columns should still be treated as binned
    for the purpose of not getting an error.
    """

    def __init__(self, source_column, result_column):
        super().__init__(source_column, result_column, "NoBin")


class Bins(object):
    """
    A class to manage the binning of columns, using the provided binning methods.
    Automatically selects the appropriate binning method based on the column type.
    """

    @staticmethod
    def default_binning_method(source_column: Series, result_column: Series) -> list:
        """
        The default binning method, which returns an empty list.
        Arguments can be ignored in this current implementation, but are kept for future use.
        :param source_column:
        :param result_column:
        :return:
        """
        return []

    BIN_SIZES = [5, 10]
    USER_BINNING_METHOD = default_binning_method
    ONLY_USER_BINS = False

    def __init__(self, source_column: Series, result_column: Series, bins_count: int):
        """
        :param source_column: The source column to bin.
        :param result_column: The result (after an operation on the source) column to bin.
        :param bins_count: The number of bins to create.
        """
        self.max_bin_count = bins_count
        self.bins = list(Bins.USER_BINNING_METHOD(source_column, result_column))
        gb = False
        if Bins.ONLY_USER_BINS:
            return

        # If the source column is None, we can infer that the result column is the result of a groupby operation.
        # Additionally, if there are multi-level indexes, we return
        if source_column is None:
            # GroupBy
            gb = True
            self.bins += self.get_multi_index_bins(source_column, result_column, bins_count)
            if len(self.bins) > 0:
                return

        # If the source column is empty and the result column is empty, return
        if source_column is not None and len(source_column) == 0 and len(result_column) == 0:
            return

        if utils.is_numeric(result_column):
            # If the result column is numeric but is the result of a groupby operation, or if the source and result column
            # have less than 15 unique values, bin the columns using the categorical binning method,
            # despite the column being numeric
            if gb or (source_column.value_counts().shape[0] < 15 and result_column.value_counts().shape[0] < 15):
                self.bins += self.bin_categorical(source_column, result_column, bins_count)
            # Otherwise, bin the columns using the numeric binning method
            self.bins += self.bin_numeric(source_column, result_column, bins_count)
        else:
            self.bins += self.bin_categorical(source_column, result_column, bins_count)

    @staticmethod
    def register_binning_method(method, use_only_user_bins=False) -> None:
        """
        Register a binning method to be used by the Bins class.
        :param method: The binning method to register.
        :param use_only_user_bins: Whether to use only the user-defined binning method.
        """
        Bins.USER_BINNING_METHOD = method
        Bins.ONLY_USER_BINS = use_only_user_bins

    @staticmethod
    def bin_numeric(source_col, res_col, size) -> List[NumericBin]:
        """
        Bin numeric columns using the NumericBin class.

        This method creates bins for the given size, and returns a list of the created bins.

        :param source_col: The source column to bin.
        :param res_col: The result column to bin.
        :param size: The number of bins to create.
        :return: A list containing the created bins.
        """
        numeric_bins = []
        for bin_count in Bins.BIN_SIZES:
            if bin_count > size:
                break
            numeric_bins.append(NumericBin(source_col, res_col, bin_count))

        return numeric_bins

    @staticmethod
    def set_numeric_bin_sizes(bin_size_list: list):
        """

        :param bin_size_list:
        :return:
        """
        Bins.BIN_SIZES = bin_size_list

    @staticmethod
    def bin_categorical(source_col: Series, res_col: Series, size: int) -> List[CategoricalBin]:
        """
        Bin categorical columns using the CategoricalBin class.
        :param source_col: The source column to bin.
        :param res_col: The result column to bin.
        :param size: The number of bins to create.
        :return: A list containing the binned column.
        """
        return [CategoricalBin(source_col, res_col, size)]

    @staticmethod
    def get_multi_index_bins(source_col, res_col, size):
        """
        Get bins for multi-index columns.

        This method checks if the result column has a multi-index and creates bins for each level of the index
        that meets the size criteria.

        :param source_col: The source column to bin.
        :param res_col: The result column to bin.
        :param size: The maximum size of the bins.
        :return: A list of MultiIndexBin objects.
        """
        if type(res_col.index) is not pd.MultiIndex:
            return []

        shortest_level = min(res_col.index.levels, key=len)

        if len(shortest_level) > size:
            # above size limit
            return []

        bins_candidates = []
        for level_index, level in enumerate(res_col.index.levels):
            if len(res_col.index.levels) > 1 and level_index == 0:
                continue
            if 1 < len(level) <= size:
                bins_candidates.append(MultiIndexBin(source_col, res_col, level_index))

        return bins_candidates



import result_comparison_functions as rcf

test_fail_explanations = {
    "Correlated attributes comparison test": "The list of correlated attributes in the re-produced results does not match the list of correlated attributes in the saved results. There is likely an error in the method get_correlated_attributes of the Filter operation.",
    "Measure scores comparison test": "The measure scores in the re-produced results do not match the measure scores in the saved results. There is likely an error in the method calc_measure_internal of the tested measure.",
    "Score dicts comparison test": "The score dictionaries in the re-produced results do not match the score dictionaries in the saved results. There is likely an error in the method calc_measure_internal of the tested measure, or in calc_measure in BaseMeasure.",
    "Influence values comparison test": "The influence values in the re-produced results do not match the influence values in the saved results. If the measure scores and score dictionaries are correct, there is likely an error in the method calc_influence_col in the tested measure.",
    "Significance values comparison test": "The significance values in the re-produced results do not match the significance values in the saved results. If the influence values test passed, then there is likely an error in the method get_significance in BaseMeasure.",
    "Results dataframe comparison test": "The results dataframes in the re-produced results do not match the results dataframes in the saved results. If all other tests have passed, then there is likely an error in the build_explanation method of the tested measure",
    "Column names comparison test": "The column names in the re-produced results do not match the column names in the saved results. There is likely an error in the method _get_column_names of the GroupBy operation.",
    "One to many attributes comparison test": "The one to many attributes in the re-produced results do not match the one to many attributes in the saved results. There is likely an error in the method get_one_to_many_attributes of the GroupBy operation.",
}

test_funcs = {
    "Correlated attributes comparison test": {
        "func": rcf.compare_correlated_attributes,
        "attribute_name": "correlated_attributes",
        "require_duplicate_fix": False
    },
    "Column names comparison test": {
        "func": rcf.compare_column_names,
        "attribute_name": "column_names",
        "require_duplicate_fix": False
    },
    "One to many attributes comparison test": {
        "func": rcf.compare_one_to_many_attributes,
        "attribute_name": "one_to_many_attributes",
        "require_duplicate_fix": False
    },
    "Measure scores comparison test": {
        "func": rcf.compare_measure_scores,
        "attribute_name": "measure_scores",
        "require_duplicate_fix": False
    },
    "Score dicts comparison test": {
        "func": rcf.compare_score_dicts,
        "attribute_name": "score_dict",
        "require_duplicate_fix": False
    },
    "Influence values comparison test": {
        "func": rcf.compare_influence_vals,
        "attribute_name": "influence_vals",
        "require_duplicate_fix": True
    },
    "Significance values comparison test": {
        "func": rcf.compare_significance_vals,
        "attribute_name": "significance_vals",
        "require_duplicate_fix": True
    },
    "Results dataframe comparison test": {
        "func": rcf.compare_results,
        "attribute_name": "saved_results",
        "require_duplicate_fix": False
    },
}
